#ifndef LIP_H
#define LIP_H
#include <config.h>
#include <mpi.h>
#include <stdio.h>

typedef int LIP_Handle;
typedef LIP_Handle LIP_Indextype; /* has to be fast to dereference*/

/* to do:
- NULL handles for all types -> 0 (others start from 1 and are decremented
  befor use)
- predefined handles
*/

/* Rules for flag arguments:
can be combined with | and tested with & in C/C++ or special intrinsic
functions (IOR) in Fortran 90 or `+' in Fortran 77. For quering flags use
`SUBROUTINE BIT_QUERY()' from `Querying File Parameters' section of MPI-2.
*/
/* typedef int LIP_Flag;*/ /* ??? (integral type with sufficient size) */

/*
flag: -> 2 flags
 LIP_USE_IND1
 LIP_USE_IND2
maybe more general indexing ?
 LIP_C_IND1
 LIP_C_IND2
 LIP_FORTRAN_IND1
 LIP_FORTRAN_IND2
typedef void LIP_User_function(int *indarr1, void *invec, int *indarr2,
  void *inoutvec, int *len, MPI_Datatype *datatype, int *flag);
*/
/**/
typedef void (*LIP_User_function)(void *iin, LIP_Indextype tin, void *invec,
  void *iinout, LIP_Indextype tinout, void *inoutvec, int *count,
  MPI_Datatype *datatype);

typedef LIP_Handle LIP_Op;
/* predefined operations */
#define LIP_OP_NULL   0
#define LIP_OP_SUM    1
#define LIP_OP_FIRST  (LIP_OP_SUM + 1)

/* operation is always associative and commutative(no order can be specified in
 * MPI collective calls and MPI_Reduce() is inefficient for irregular access)
 */
int LIP_Op_create(LIP_User_function *function, LIP_Op *op);
int LIP_Op_free(LIP_Op *op);

#define LIP_SCHEDULE_NULL 0
typedef LIP_Handle LIP_Schedule;
int LIP_Schedule_create(int size, LIP_Schedule *schedule); /* size of local data array */
int LIP_Schedule_commit(LIP_Schedule *schedule); /* makes ready for communication */
int LIP_Schedule_buffer_attach(LIP_Schedule schedule, void *buf, int size); /* in bytes(maybe in buffer elements?) */
int LIP_Schedule_buffer_detach(LIP_Schedule schedule, void *buf, int *size); /* in bytes */
int LIP_Schedule_free(LIP_Schedule *schedule);
/*
 LIP_EXPORT_SIZE
 LIP_DATA_COUNT
 LIP_IMPORT_SIZE
 LIP_COMM_BUF_SIZE(for buffer_attach)
 LIP_SCHEDULE_INFO_SIZE - constant size of info array
Maybe like MPI_Status - struct in C, array in Fortran?
*/
int LIP_Schedule_get_info(LIP_Schedule schedule, int *info, int count);

typedef LIP_Handle LIP_Datamap;
/* predefined data maps (no parameter information required) */
#define LIP_DATAMAP_NULL   0
#define LIP_CYCLIC 1 /* equivalent to HPF 2.0 CYCLIC distribution
                                (i.e.  CYCLIC(1)) */
#define LIP_DATAMAP_FIRST  (LIP_CYCLIC + 1) /* first not-predefined
                                                       datamap */

/* types of distribution */
#define LIP_DATAMAP_BLOCK     0 /* HPF 2.0: BLOCK(m) */
#define LIP_DATAMAP_CYCLIC    1 /* HPF 2.0: CYCLIC(m) */
#define LIP_DATAMAP_INDIRECT  2 /* HPF 2.0: INDIRECT */
#define LIP_DATAMAP_GEN_BLOCK 3 /* HPF 2.0: GEN_BLOCK */

typedef LIP_Handle LIP_Indexinfo;

#define LIP_INDEXINFO_NULL 0


/*
flag:
 LIP_UNIQUE
 LIP_OVERWRITE
 LIP_UPDATE_INFO
  type of problem(??):
 LIP_MULTIPHASE
 LIP_ADAPTIVE
*/
#define MAX_SCHEDULE 100
int LIP_Localize(LIP_Datamap datamap, void *iglobal, LIP_Indextype igtype,
  void *ilocal, LIP_Indextype iltype, int count, int flags,
  LIP_Schedule *schedule, LIP_Indexinfo *info);

/* similar to MPI_Gather() where current node is a root */
int LIP_Gather(void *sendbuf, void *recvbuf, MPI_Datatype datatype,
  LIP_Schedule schedule);
/* similar to MPI_Scatter() where current node is a root */
int LIP_Scatter(void *sendbuf, void *recvbuf, MPI_Datatype datatype,
  LIP_Schedule schedule, LIP_Op op);

#define LIP_IOBUFMAP_NULL 0
typedef LIP_Handle LIP_IOBufmap;
int LIP_IOBufmap_create(int size, LIP_IOBufmap *bmap);
int LIP_IOBufmap_free(LIP_IOBufmap *bmap);
/*LIP_IObufmap_add(int size, LIP_IObufmap bmap);*/ /* for optimization */
/*
flag: - two phase IO buffer exchange
 LIP_DATA_ALL (= OLD+NEW) ?
 LIP_DATA_OLD
 LIP_DATA_NEW
*/
#define MAX_IOBUFMAP 100
int LIP_IOBufmap_get_datatype(LIP_IOBufmap bmap, MPI_Datatype datatype,
  MPI_Datatype *memtype, MPI_Datatype *filetype, int flags);

/* flags:
 LIP_UNIQUE
 LIP_OVERWRITE
*/
int LIP_OOC_localize(void *ilocal, LIP_Indextype iltype, void *itlocal,
  LIP_Indextype ittype, int count, int flags, void *buf, MPI_Datatype datatype,
  LIP_IOBufmap *bmap, LIP_Schedule *schedule);


/*
flags:
*/
#define  LIP_LOCAL 0
#define  LIP_NONLOCAL 1
#define  LIP_OOC 0x02
/*
type:
FUNCTION (?) -> function_attach()

change: added desctype
*/
#define GHOST_AREA_RATE 2
#define MAX_DATAMAP 100
int LIP_Datamap_create(int type, int global, void *desc, LIP_Indextype desctype,
  int count, int flags, LIP_Datamap *datamap);


/* flags: 
// 0 - delete files with ooc translation table
// LIP_DATAMAP_STORE - close files but do not delete from disk
*/
#define LIP_DATAMAP_STORE 1
int LIP_Datamap_free(LIP_Datamap *datamap, int flags);

int LIP_copy_files (MPI_File data_file,
                   MPI_File data_out_file,
                   MPI_Offset offset,int count,
                   int page_size,MPI_Datatype datatype);

/* zmiana !!!! */
#define LIP_COPY 1
/*-processors table is copied - requires more time
and memory,
but one can change proc array afterwords*/
#define LIP_ATTACH 0
/*a pointer to the processors table is copied-
one should not change procs array afterwords 
*/

 
int LIP_Datamap_ic_ttable_copy(LIP_Datamap datamap,
  void *procs, LIP_Indextype procstype,
  void *local_adr,LIP_Indextype localtype,
  int ttable_size,int flag);

int LIP_Datamap_ooc_to_local(
  LIP_Datamap datamap,
  char* local_dir);

int LIP_Datamap_ooc_ttable_attach(
  LIP_Datamap datamap,
  char * proc_fname,
  char * local_adr_fname,
  int ttable_size,
  int ttable_membuf_size);

int LIP_Datamap_ooc_ttable_get_files(
  LIP_Datamap datamap,
  char * procs_fname,
  char * local_adr_fname);


int LIP_Datamap_ic_ttable_detach(LIP_Datamap datamap,
 int** procs,int** local_adr,int* ttable_size);

int LIP_Datamap_ic_ttable_free(LIP_Datamap datamap);


int LIP_Datamap_ic_init_table_with_proc(LIP_Datamap datamap,void*proc,
  LIP_Indextype procstype,
  int ttable_size, int flags );

/*
predefined(type + indexing???what if C int with Fortran indexing):
support for `long long' (64 bits)?
 LIP_INTEGER
 LIP_LONG_LONG
 LIP_LONG
 LIP_INT
 LIP_SHORT ?
 LIP_CHAR ???
*/

/*
general integer array:
a[i] -> base_address + offset + ceil(i / count) * bsize + shifts[i (mod count)]
 
flags:
 LIP_INDEXTYPE_C, LIP_INDEXTYPE_FORTRAN
indexing C, FORTAN, maybe more general???
*/
int LIP_Indextype_block(int offset, int blksize, int *blkdispl, int displcount,
    int flags, LIP_Indextype *itype);
/*
Many arrays together + multiple levels of indirection.
int LIP_Indextype_commit(LIP_Indextype *itype);
change: LIP_Indextype_addresses -> LIP_Indextype_multiple
*/
int LIP_Indextype_multiple(MPI_Aint *displ, LIP_Indextype *itypes, int count,
  int mlevel, LIP_Indextype *itype);
int LIP_Indextype_free(LIP_Indextype *itype);

int LIP_Indexinfo_create(LIP_Indexinfo *iinfo);
/*
flags:
 LIP_INDEXINFO_UNIQUE - find globally unique
 LIP_INDEXINFO_NODE_UNIQUE - find unique on each node
 LIP_INDEXINFO_NODE_COUNT - find count for each node
*/
int LIP_Indexinfo_analyze(LIP_Datamap dmap, void *indarr, LIP_Indextype itype,
  int count, int flags, LIP_Indexinfo *info);
  int LIP_Indexinfo_set(int count, int ucount, int *inode, int *iunique,
  LIP_Indexinfo *info);
int LIP_Indexinfo_get(LIP_Indexinfo info, int *count, int *ucount, int *inode,
  int *iunique);
int LIP_Indexinfo_free(LIP_Indexinfo *info);

int LIP_Setup(int *nodes, int count); /* count=0 -> all nodes from MPI_COMM_WORLD(or UNIVERSE?) */
int LIP_Exit(void);
int LIP_Rank(int *node);
int LIP_Size(int *nodes);

/*max length of error string*/
#define LIP_MAX_ERROR_STR_LENGTH 200

/*function to get string associated with an error number*/

int LIP_Error_mesg(int error_code,char* error_str,int *length);
/*procedure name to report errors*/

int LIP_Error_quit(char*pname,char* error_mesg,int error_code);
int LIP_Error_ret(char*pname,char* error_mesg,int error_code);
int LIP_Assert(void* ptr,char*pname,char* error_mesg,int error_code);


/* return values */
#define LIP_SUCCESS 0
#define LIP_ERROR   1
#define LIP_OUT_OF_MEMORY 2
#define LIP_NOT_SUPPORTED 3
#define LIP_INVALID_ARGUMENT 4
#define LIP_TOO_MANY_HANDLES 5
#define LIP_NOT_COMMITED 6

/* predefined indexing schemes */
#define LIP_INDEXTYPE_NULL    0
#define LIP_INDEXTYPE_INT     1 /* C indexing (starting with 0) using `int' */
#define LIP_INDEXTYPE_FIRST   (LIP_INDEXTYPE_INT + 1) /* first not-predefined
                                                         indextype */

#define LIP_DATA_NEW 1
#define LIP_DATA_OLD 2
#define LIP_DATA_ALL 4

#define GHOST_AREA_RATE 2

int LIP_create_hilbert_distribution_in_core
 (LIP_Datamap datamap,LIP_Datamap *new_datamap,double *coX,double *coY,
 double *coZ,
  double* vertex,double edge,double minedge,int sample_count);



int LIP_create_hilbert_distribution_ooc
 (LIP_Datamap datamap,  LIP_Datamap * new_datamap, MPI_File XFILE,MPI_File YFILE,
 MPI_File ZFILE,int count,int page_size,int sample_count,
  void* vertex,double edge,double minedge);
int LIP_remap_in_core(LIP_Datamap olddm,LIP_Datamap newdm ,LIP_Schedule* sch);
int LIP_remap_ooc (LIP_Datamap datamap,
                   LIP_Datamap * new_datamap,
                   MPI_File data_file,
                   MPI_File data_out_file, 
                   int count, 
                   int page_size,
                   MPI_Datatype datatype,
                   int flags);

/*flags for LIP_remap_ooc function*/

/* local indices part of translation table is created*/
#define LIP_CREATE_LOCAL 1 
/* data from data_file is moved to data_out_file according
                            to new mapping */
#define LIP_MOVE_DATA 2 


#endif /* #ifndef LIP_H */
  
