/* -*-Java-*- */

package pl.edu.agh.icsr.lip;

import pl.edu.agh.icsr.janet.tools.*;

interface DataDistribution {
    int getGlobalSize();
    int getType();
    int getDescType();
    Object getDesc();
}

public class Datamap extends DataEncapsulator {
native "C" {
#include <lip.h>
#include <mpi.h>
#include <mpio.h>
}
    public static class BlockDistribution implements DataDistribution {
	int gsize, lsize;
	public BlockDistribution(int gsize, int lsize) {
	    this.gsize = gsize;
	    this.lsize = lsize;
	}

	public int getGlobalSize() {
	    return gsize;
	}

	public int getType() { return Datamap.BLOCK; }

	public int getDescType() { return LIP.INDEXTYPE_INT; }
	
	public Object getDesc() {
	    return new VInt(new int[] { lsize });
	}
    }

    private static long DATAMAP_NULL = 0;
  
    private static final int BLOCK     = 0; /* HPF 2.0: BLOCK(m) */
    private static final int CYCLIC    = 1; /* HPF 2.0: CYCLIC(m) */
    private static final int INDIRECT  = 2; /* HPF 2.0: INDIRECT */
    private static final int GEN_BLOCK = 3; /* HPF 2.0: GEN_BLOCK */


    public long getNullValue() { return DATAMAP_NULL; }

    public Datamap(DataDistribution distrib, int count, int flags) {
	init(distrib.getType(), distrib.getGlobalSize(),
	     distrib.getDesc(), distrib.getDescType(), count, flags);
    }

    private native "C" void init(int type, int global, Object desc, int desctype, int count,
				 int flags) {
	LIP_Datamap d;
	int result;
	`VInt vintdesc = (VInt)desc;` /* currently all distibution types use VInt */
	    result = LIP_Datamap_create(`type`, `global`, 
					&`#&vintdesc.getBaseArray()`[`vintdesc.getOffset()`],
					LIP_INDEXTYPE_INT, `count`, `flags`, &d);
        `this.data = #(d)`;
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    public native "C" void free(int flags) {
	LIP_Datamap d = (LIP_Datamap)`this.data`;
	int result = LIP_Datamap_free(&d, `flags`);
	`this.data = #(d);`
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    protected void freeNative() {
	if (data != getNullValue()) free(0);
    }
}

