/* -*-Java-*- */

package pl.edu.agh.icsr.lip;

import mpi.*;
import pl.edu.agh.icsr.janet.tools.*;

interface Indexer {
    Object getIndexData();
    long getIndexType();
    int getCount();
}
    
public class LIP {
native "C" {
#include <lip.h>
#include <mpi.h>
#include <mpio.h>
#include <debug.h>
}

//     interface Indexinfo {
// 	long getIndexinfo();
//     }

//    private final static int LIP_INDEXINFO_NULL = 0;
    
    static final int INDEXTYPE_NULL = 0;
    static final int INDEXTYPE_INT = 1;
    static final int INDEXTYPE_FIRST = (INDEXTYPE_INT + 1);

    public final static int DATA_NEW = 1;
    public final static int DATA_OLD = 2;
    public final static int DATA_ALL = 4;

    public final static int OP_NULL = 0;
    public final static int OP_SUM  = 1;
    public final static int OP_FIRST = (OP_SUM + 1);



//     public static Indexinfo INDEXINFO_NULL = new Indexinfo() {
//         public long getIndexinfo() { return LIP_INDEXINFO_NULL;	}
//     };

    public static class ContIntIndexer implements Indexer {
	VInt indcs;

	public ContIntIndexer(VInt indcs) {
	    this.indcs = new VInt(indcs.getBaseArray()).
		subArray(indcs.getOffset(), indcs.getLength());
	}

	public ContIntIndexer(int[] indcs) {
	    this.indcs = new VInt(indcs);
	}

	public ContIntIndexer(int[] indcs, int offset, int length) {
	    this.indcs = new VInt(indcs).subArray(offset, length);
	}

	public Object getIndexData() {
	    return indcs;
	}

	public long getIndexType() {
	    return INDEXTYPE_INT;
	}

	public int getCount() {
	    return indcs.getLength();
	}
    }

    public static native "C" void Setup(int[] nodes, int count) {
	int* nodes;
	int result;
	if (`nodes == null`) {
	    nodes = 0;
	} else {
	    nodes = `#&nodes`;
	}
	result = LIP_Setup(nodes, `count`);
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }
  
    public static native "C" void Exit() {
	int result = LIP_Exit();
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    public static native "C" int Rank() {
	int node;
	int result = LIP_Rank(&node);
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
	return node;
    }
  
    public static native "C" int Size() {
	int nodes;
	int result = LIP_Size(&nodes);
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
	return nodes;
    }

    public static native "C" void DEBUG_create_file() {
	int result = LIP_DEBUG_create_file();
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    public static void Localize(Datamap datamap, Indexer iglobal, Indexer ilocal,
				int flags, Schedule schedule) {
	int count = iglobal.getCount();
	if (count != ilocal.getCount()) {
	    throw new LipException("Global and local indexers do not have the same " +
				   "number of indices");
	}
	nlocalize(datamap, 
		  iglobal.getIndexData(), iglobal.getIndexType(),
		  ilocal.getIndexData(), ilocal.getIndexType(),
		  count, flags, schedule);
    }

    private static native "C" void nlocalize(Datamap datamap, Object igdata, long igtype,
					     Object ildata, long iltype, int count, int flags,
					     Schedule schedule) {
	LIP_Schedule s = (LIP_Schedule)`schedule.data`;
	int result;
	`VInt vigdata = (VInt)igdata;` /* all currently supported return VInt */
	`VInt vildata = (VInt)ildata;` /* all currently supported return VInt */
	result = LIP_Localize((LIP_Datamap)`datamap.data`,
			      &`#&vigdata.getBaseArray()`[`vigdata.getOffset()`],
			      `igtype`,
			      &`#&vildata.getBaseArray()`[`vildata.getOffset()`],
			      `iltype`,
			      `count`, `flags`,
			      &s, LIP_INDEXINFO_NULL);
	`schedule.data = #(s);`
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    public static void OOC_Localize(Indexer ilocal, Indexer itlocal,
				    int flags, VDouble buf,
				    mpi.Datatype datatype,
				    IOBufmap bmap, Schedule schedule) {
	int count = ilocal.getCount();
	if (count != itlocal.getCount()) {
	    throw new LipException("local and tlocal indexers do not have the same " +
				   "number of indices");
	}
	
	nOOC_localize(ilocal.getIndexData(), ilocal.getIndexType(),
		      itlocal.getIndexData(), itlocal.getIndexType(),
		      count, flags, buf, datatype, bmap, schedule);
    }

    private static native "C" void nOOC_localize(Object ilocal, long iltype,
						 Object itlocal, long ittype,
						 int count, int flags,
						 VDouble buf,
						 Datatype datatype,
						 IOBufmap bmap, Schedule schedule) {
	    
	LIP_Schedule s = (LIP_Schedule)`schedule.data`;
	LIP_IOBufmap b = (LIP_IOBufmap)`bmap.data`;
	int result;
	`VInt vilocal = (VInt)ilocal;` /* all currently supported return VInt */
	`VInt vitlocal = (VInt)itlocal;` /* all currently supported return VInt */

        result = LIP_OOC_localize(&`#&vilocal.getBaseArray()`[`vilocal.getOffset()`],
                                  `iltype`,
                                  &`#&vitlocal.getBaseArray()`[`vitlocal.getOffset()`],
                                  `ittype`,
	                          `count`, `flags`,
                                  &`#&buf.getBaseArray()`[`buf.getOffset()`],
	                          (MPI_Datatype)`datatype.handle`,
	                          &b, &s);

	`schedule.data = #(s);`
	`bmap.data = #(b);`

	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }

    public static native "C" void Gather(VDouble sendbuf, VDouble recvbuf, Datatype datatype,
					Schedule schedule) {
	int result = LIP_Gather(&`#&sendbuf.getBaseArray()`[`sendbuf.getOffset()`],
	&`#&recvbuf.getBaseArray()`[`recvbuf.getOffset()`],
	(MPI_Datatype)`datatype.handle`,
	(LIP_Schedule)`schedule.data`);
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }
  
    public static native "C" void Scatter(VDouble sendbuf, VDouble recvbuf, Datatype datatype,
					  Schedule schedule, int op) {
	int result = LIP_Scatter(&`#&sendbuf.getBaseArray()`[`sendbuf.getOffset()`],
	&`#&recvbuf.getBaseArray()`[`recvbuf.getOffset()`],
	(MPI_Datatype)`datatype.handle`,
	(LIP_Schedule)`schedule.data`,
	`op`);
	if (result != LIP_SUCCESS) `throw new LipException(#(result));`
    }
  
  
}
