/*
 LIP library
 comm.c
 contains communication functions
 created by Piotr Luszczek
 modified by Maciej Malawski, Katarzyna Zajac
*/

#include <stdlib.h>

#include <lipimpl.h>
#include <debug.h>
static int
LIP_Gather_d(double *sendbuf, double *recvbuf, LIP_ISchedule *is)
{
  int i;
  int allocated_buffer = 0;
  char* pname="LIP_Gather_d";
  double *rsendbuf; /* real send buffer */

  int rsendbuf_size =
      is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];

  if (NULL == is->user_buf)
  {

    rsendbuf = malloc( (sizeof *rsendbuf) * rsendbuf_size );
    
    LIP_Assert(rsendbuf,pname,NULL,LIP_OUT_OF_MEMORY);
    allocated_buffer = 1;

  }
  else
    rsendbuf = is->user_buf; /* use user-supplied communication buffer */

  /* for all the data elements that are supposed to be exported */
  for (i = rsendbuf_size - 1 ; i >= 0 ; i--)
    rsendbuf[i] = sendbuf[ is->elocations[i] ];

  MPI_Alltoallv( rsendbuf, is->esizes, is->edispls, MPI_DOUBLE, recvbuf,
    is->isizes, is->idispls, MPI_DOUBLE, LIP_env.comm );

  if (allocated_buffer)
    free( rsendbuf );

  return LIP_SUCCESS;
} /* LIP_Gather_d */

static int
LIP_Gather_i(int *sendbuf, int *recvbuf, LIP_ISchedule *is)
{
  int i,kk;
  int allocated_buffer = 0;
  char* pname="LIP_Gather_i";
 
  int *rsendbuf; /* real send buffer */
  int rsendbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];


  if (NULL == is->user_buf)
  {

    rsendbuf = malloc( (sizeof *rsendbuf) * rsendbuf_size );
    LIP_Assert(rsendbuf,pname,NULL,LIP_OUT_OF_MEMORY);

    allocated_buffer = 1;
  }
  else
    rsendbuf = is->user_buf; /* use user-supplied communication buffer */

  /* for all the data elements that are supposed to be exported */
  for (i = rsendbuf_size - 1 ; i >= 0 ; i--)
    rsendbuf[i] = sendbuf[ is->elocations[i] ];

#ifdef DEBUG
  fprintf(debf,"Scheduler's fields: esizes, edispls,isizes,idispl:\n");
   
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->esizes[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->edispls[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->isizes[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->idispls[kk]);
  }
  fprintf(debf,"\n");
 
#endif
  
  MPI_Alltoallv( rsendbuf, is->esizes, is->edispls, MPI_INT, recvbuf,
    is->isizes, is->idispls, MPI_INT, LIP_env.comm );
#ifdef DEBUG
 fprintf(debf,"przeslalem\n");
#endif
  if (allocated_buffer)
    free( rsendbuf );

  return LIP_SUCCESS;
} /* LIP_Gather_i */
static int
LIP_Gather_dt(void *sentbufg, void *recvbufg, LIP_ISchedule *is,
MPI_Datatype type)
{
  int i,j,kk;
  int allocated_buffer = 0;
  char* pname="LIP_Gather_dt";
  char* sendbuf=(char*)sentbufg;
  char* recvbuf=(char*)recvbufg;
  int sizet;
  char* rsendbuf; /* real send buffer */
  int rsendbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];
  MPI_Type_size(type,&sizet);
  
  if (NULL == is->user_buf)
  {

    rsendbuf = malloc( sizet * rsendbuf_size );
    LIP_Assert(rsendbuf,pname,NULL,LIP_OUT_OF_MEMORY);

    allocated_buffer = 1;
  }
  else
    rsendbuf = is->user_buf; /* use user-supplied communication buffer */

  /* for all the data elements that are supposed to be exported */
  for (i = rsendbuf_size - 1 ; i >= 0 ; i--){
   for(j=0;j<sizet;j++){
    rsendbuf[i*sizet+j] = sendbuf[ sizet*(is->elocations[i])+j ];
   }
  }
#ifdef DEBUG
  fprintf(debf,"Scheduler's fields: esizes, edispls,isizes,idispl:\n");
   
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->esizes[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->edispls[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->isizes[kk]);
  }
  fprintf(debf,"\n");
  for(kk=0;kk<LIP_env.size;kk++){
   fprintf(debf,"%d ",is->idispls[kk]);
  }
  fprintf(debf,"\n");
 
#endif
  MPI_Alltoallv( (void*)rsendbuf, is->esizes, is->edispls, type, recvbuf, is->isizes,
   is->idispls, type, LIP_env.comm );
#ifdef DEBUG
 fprintf(debf,"przeslalem\n");
#endif
  if (allocated_buffer)
    free( rsendbuf );

  return LIP_SUCCESS;
} /* LIP_Gather_i */

static int LIP_Scatter_dt(void *sendbufg, void *recvbufg, LIP_ISchedule 
*is,MPI_Datatype type )
{
  int allocated_buffer = 0;
  char* pname="LIP_Scatter_dt";
  char* sendbuf=(char*)sendbufg;
  char* recvbuf=(char*)recvbufg;
  int sizet,i,j;
  char* rrecvbuf; /* real receive buffer */
  int rrecvbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];
  MPI_Type_size(type,&sizet);
  
  if (NULL == is->user_buf)
  {

    rrecvbuf = malloc( sizet * rrecvbuf_size );
    LIP_Assert(rrecvbuf,pname,NULL,LIP_OUT_OF_MEMORY);

    allocated_buffer = 1;
  }
  else
    rrecvbuf = is->user_buf; /* use user-supplied communication buffer */


  MPI_Alltoallv( sendbuf, is->isizes, is->idispls, type, 
(void*)rrecvbuf,
    is->esizes, is->edispls, type, LIP_env.comm );

  /* for all the data elements that were brought back */
  for (i = rrecvbuf_size - 1 ; i >= 0 ; i--)
    for(j=0;j<sizet;j++)
    recvbuf[ sizet*(is->elocations[i])+j ] = rrecvbuf[i*sizet+j];

  if (allocated_buffer)
    free( rrecvbuf );

  return LIP_SUCCESS;
}

static int
LIP_Scatter_d
(double *sendbuf, double *recvbuf, LIP_ISchedule *is)
{
  int i;
  int allocated_buffer = 0;
  char* pname="LIP_Scatter_d";
  double *rrecvbuf; /* real receive buffer */
  int rrecvbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];

  if (NULL == is->user_buf)
  {

    rrecvbuf = malloc( (sizeof *rrecvbuf) * rrecvbuf_size );
    LIP_Assert(rrecvbuf,pname,NULL,LIP_OUT_OF_MEMORY);
    
    allocated_buffer = 1;
  }
  else
    rrecvbuf = is->user_buf; /* use user-supplied communication buffer */

  MPI_Alltoallv( sendbuf, is->isizes, is->idispls, MPI_DOUBLE, rrecvbuf,
    is->esizes, is->edispls, MPI_DOUBLE, LIP_env.comm );

  /* for all the data elements that were brought back */
  for (i = rrecvbuf_size - 1 ; i >= 0 ; i--)
    recvbuf[ is->elocations[i] ] = rrecvbuf[i];

  if (allocated_buffer)
    free( rrecvbuf );

  return LIP_SUCCESS;
} /* LIP_Scatter_d */

static int
LIP_Scatter_d_sum(double *sendbuf, double *recvbuf, LIP_ISchedule *is)
{
  int i;
  int allocated_buffer = 0;
  char* pname="LIP_Scatter_d_sum";
  double *rrecvbuf; /* real receive buffer */
  int rrecvbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];

  if (NULL == is->user_buf)
  {
    
    rrecvbuf = malloc( (sizeof *rrecvbuf) * rrecvbuf_size );
    LIP_Assert(rrecvbuf,pname,NULL,LIP_OUT_OF_MEMORY);

    allocated_buffer = 1;
  }
  else
    rrecvbuf = is->user_buf; /* use user-supplied communication buffer */

  MPI_Alltoallv( sendbuf, is->isizes, is->idispls, MPI_DOUBLE, rrecvbuf,
    is->esizes, is->edispls, MPI_DOUBLE, LIP_env.comm );

  /* for all the data elements that were brought back */
  for (i = rrecvbuf_size - 1 ; i >= 0 ; i--)
    recvbuf[ is->elocations[i] ] += rrecvbuf[i];

  if (allocated_buffer)
    free( rrecvbuf );

  return LIP_SUCCESS;
} /* LIP_Scatter_d_sum */

static int
LIP_Scatter_i_sum(int *sendbuf, int *recvbuf, LIP_ISchedule *is)
{
  int i;
  int allocated_buffer = 0;
  char* pname="LIP_Scatter_i_sum";
  int *rrecvbuf; /* real receive buffer */
  int rrecvbuf_size =
    is->edispls[LIP_env.size - 1] + is->esizes[LIP_env.size - 1];

  if (NULL == is->user_buf)
  {
    
    rrecvbuf = malloc( (sizeof *rrecvbuf) * rrecvbuf_size );
    LIP_Assert(rrecvbuf,pname,NULL,LIP_OUT_OF_MEMORY);

    allocated_buffer = 1;
  }
  else
    rrecvbuf = is->user_buf; /* use user-supplied communication buffer */

  MPI_Alltoallv( sendbuf, is->isizes, is->idispls, MPI_INT, rrecvbuf,
    is->esizes, is->edispls, MPI_INT, LIP_env.comm );

  /* for all the data elements that were brought back */
  for (i = rrecvbuf_size - 1 ; i >= 0 ; i--)
    recvbuf[ is->elocations[i] ] += rrecvbuf[i];

  if (allocated_buffer)
    free( rrecvbuf );

  return LIP_SUCCESS;
} /* LIP_Scatter_i_sum */

/* ---------------------------------------------------------------------- */

int
LIP_Gather(void *sendbuf, void *recvbuf, MPI_Datatype datatype,
  LIP_Schedule schedule)
{
  char* pname="LIP_Gather";
    
  LIP_ISchedule *is = LIP_Schedule_convert( schedule );
  /* if there is only one node */
  if (1 == LIP_env.size)
    return LIP_SUCCESS;

  LIP_Assert(is,pname,"schedule handle",LIP_INVALID_ARGUMENT);


  /* if schedule wasn't committed */
  if (! is->committed){
    LIP_Error_quit(pname,"schedule",LIP_NOT_COMMITED);
  }
  
  if (MPI_DOUBLE == datatype)
     return LIP_Gather_d( sendbuf, recvbuf, is );
  if (MPI_INT == datatype)
     return LIP_Gather_i(sendbuf, recvbuf, is);
  else
    return LIP_Gather_dt(sendbuf,recvbuf,is,datatype);
/*  LIP_Error_ret(pname,"datatype of data that are going to be gathered",
LIP_NOT_SUPPORTED);
  return LIP_NOT_SUPPORTED;*/
} /* LIP_Gather */


int
LIP_Scatter(void *sendbuf, void *recvbuf, MPI_Datatype datatype,
  LIP_Schedule schedule, LIP_Op op)
{
  LIP_ISchedule *is = LIP_Schedule_convert( schedule );
  char* pname="LIP_Scatter";

  /* if there is only one node */
  if (1 == LIP_env.size)
    return LIP_SUCCESS;

  LIP_Assert(is,pname,"schedule",LIP_INVALID_ARGUMENT);
    

  /* if schedule wasn't committed */
  if (! is->committed){
    LIP_Error_quit(pname,"schedule",LIP_NOT_COMMITED);
  }
  if (MPI_DOUBLE == datatype){
    if (LIP_OP_NULL == op){
      return LIP_Scatter_d( sendbuf, recvbuf, is );
    }
    else if (LIP_OP_SUM == op)
      return LIP_Scatter_d_sum( sendbuf, recvbuf, is );
  }
  if (MPI_INT ==datatype) {  
    	 if (LIP_OP_NULL == op)
      		return LIP_Scatter_dt( sendbuf, recvbuf, is,MPI_INT );
    	 else if (LIP_OP_SUM == op)
      		return LIP_Scatter_i_sum( sendbuf, recvbuf, is);
  }
  
  if (LIP_OP_NULL == op)
      		return LIP_Scatter_dt( sendbuf, recvbuf, is,datatype );
  if (LIP_OP_SUM == op){
		LIP_Error_ret(pname,
                  "operation for datatype of data that are going to be scattered",
                  LIP_NOT_SUPPORTED);
                return LIP_NOT_SUPPORTED;
  }    
  

 /* LIP_Error_ret(pname,"datatype of data that are going to be 
scattered",LIP_NOT_SUPPORTED);
  
  return LIP_NOT_SUPPORTED;
  */
} /* LIP_Scatter */
