/*LIP Library
 file correct.c - contains functions used by partitioner
 for distribution correction
 created by Katarzyna Zajac*/
  
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include <lip.h>
#include <lipimpl.h>
#include <correct.h>
#include <debug.h>
#define DIV(j,k) ((j)/(k))
#define REM(j,k) ((j)%(k))

/*needed by quicksort*/
int sortint(const void * s,const void * d){
return ((*(int *)s)-(*(int *)d));
}

int sortLD(const unsigned long *s,const unsigned long *d){
 return ((*s)-(*d));
}

/*merge sort for unsigned long integers*/ 
int merge_UL (unsigned long* outbuf,unsigned long * buf1,int count1,
unsigned long *buf2,int count2)
{

unsigned long * ob=outbuf;
unsigned long * b1=buf1;
unsigned long * b2=buf2;
int c,c1,c2;

c=0;
c1=0;
c2=0;

 while(c<count1+count2){

  while ((*b1)<=(*b2)) {
   *ob=*b1;
   b1++;
   c1++;
   ob++;
   c++;
   
   if (c1==count1){
    while (c<count1+count2){    
     *ob=*b2;
     b2++;
     c2++;
     ob++;
     c++;
    }
    break;
   }
  }

  while ((*b1)>(*b2)&&(c<count1+count2)) {
   *ob=*b2;
   b2++;
   c2++;
   ob++;
   c++;
   if (c2==count2){
    while (c<count1+count2){    
     *ob=*b1;
     b1++;
     c1++;
     ob++;
     c++;
    }
    break;
   }
  }
 }
 return LIP_SUCCESS;
}

/*merge sort*/
int merge (void* outbuf,void * buf1,int count1,
void *buf2,int count2,MPI_Datatype datatype){
 if (MPI_UNSIGNED_LONG==datatype){
     return merge_UL((unsigned long*)outbuf,(unsigned long*)buf1,count1,
     (unsigned long*)buf2,count2);
 }
 else{
     return LIP_NOT_SUPPORTED;
 }
}

/*merge sorting by all processes*/
void* chain_merge (void *mybuffer,int count,MPI_Datatype datatype){
int n,npol,my_count,rec_count,i;
void *r_buf,*merge_buf,*temp,*my_buf;
unsigned long *poi;
char* pname="chain merge";
MPI_Status status;

if(1==LIP_env.size){
 return mybuffer;
}

n=2;
npol=1;
r_buf=NULL;
merge_buf=NULL;  
my_buf=mybuffer;
my_count=count;
  
/*processes that receive sorted sample and merge it with their own*/
while ((0==LIP_env.rank % n) && (LIP_env.rank+npol<LIP_env.size)){

#ifdef DEBUG
  fprintf(debf,"odbieram od %d \n",LIP_env.rank+npol);
#endif

  /*receiving sample*/
  MPI_Recv(&rec_count,1,MPI_INT,LIP_env.rank+npol,99,LIP_env.comm,&status);
  
  r_buf=realloc(r_buf,rec_count*sizeof(datatype));
  
  LIP_Assert(r_buf,pname,"during sorting phase 1",LIP_OUT_OF_MEMORY);
  
  MPI_Recv(r_buf,rec_count,datatype,LIP_env.rank+npol,99,LIP_env.comm,&status);

#ifdef DEBUG
  poi=(unsigned long*) r_buf;
  for(i=0;i<rec_count;i++){
      fprintf (debf,"%u ",poi[i]);
  }
  fprintf(debf,"\n");
#endif

  /*merging sample*/
  merge_buf=realloc(merge_buf,(rec_count+my_count)*sizeof(datatype));
  
  LIP_Assert(merge_buf,pname,"during sorting phase 2",LIP_OUT_OF_MEMORY);
  
  merge(merge_buf,r_buf,rec_count,my_buf,my_count,datatype);
  my_count=rec_count+my_count;

  /*changing pointers for next iteration*/
  if(my_buf==mybuffer){
     my_buf=merge_buf;
     merge_buf=NULL;
  }
  else{
     temp=merge_buf;
     merge_buf=my_buf;
     my_buf=temp;
  }

#ifdef DEBUG
  fprintf(debf,"Posortowane indeksy Hilberta po otrzymaniu \n");
  poi=(unsigned long*) my_buf;
  for(i=0;i<my_count;i++){
      fprintf (debf,"%u ",poi[i]);
  }
  fprintf(debf,"\n");
#endif

/*changing senders*/
  npol=n;    
  n=2*n;
 }

/*for those who don't have senders*/   
 while ((0!=LIP_env.rank)&&(0==LIP_env.rank % n) && (LIP_env.rank+npol>=LIP_env.size)){
  npol=n;
  n=2*n;
 }
 
/*senders*/
  
 if (npol=LIP_env.rank % n ){
   poi=(unsigned long*) my_buf;
   
#ifdef DEBUG
   fprintf(debf,"wysylam do %d  \n",LIP_env.rank-npol);
   for(i=0;i<my_count;i++){
      fprintf (debf,"%u ",poi[i]);
   }
   fprintf(debf,"\n");
#endif

/*send data*/ 
   MPI_Send(&my_count,1,MPI_INT,LIP_env.rank-npol,99,LIP_env.comm);
   MPI_Send(my_buf,my_count,datatype,LIP_env.rank-npol,99,LIP_env.comm);
   if (r_buf!=NULL) {
       r_buf=realloc(r_buf,0);
       r_buf=NULL;
  
   }
   if (merge_buf !=NULL) {
       merge_buf=realloc(merge_buf,0);
       merge_buf=NULL; 
   }
   return NULL;    
    
 }

/*free memory and return adress of buffer*/   
 if (0==LIP_env.rank){
    if (r_buf!=NULL) {
	r_buf=realloc(r_buf,0);
	r_buf=NULL;
  
    }
    if (merge_buf !=NULL) {
	merge_buf=realloc(merge_buf,0);
        merge_buf=NULL;
    }

    return my_buf;
 }
       
  
}

/*function for partly sorting elements - after calling this function
all elements are sorted in he way that
elements from bucket[sum_belongings[i]] to bucket[sum_belongings[i+1]]
are smaller then any elemnent of index greater then bucket[sum_belongings[i+1]] 
and greater then any element of index smaller then bucket[sum_belongings[i]]
based on quick sort algorythm
*/
int part_sort (int* sum_belongings,int count_bel,LIP_bucket_element* bucket,
                int count){
int i,j,left,right,k,l;
unsigned int xh;
LIP_bucket_element pomh;
int pom;
int *lefts,*rights;
int start=0;

while (0==sum_belongings[start]){
       start++;
       if (start==count_bel) {return 0;}
}

lefts=malloc(count_bel*sizeof(int));
rights=malloc(count_bel*sizeof(int));
lefts[0]=0;
rights[0]=count-1; 

for(i=1;i<count_bel;i++){
 lefts[i]=0;
 rights[i]=count-1;
}

if (sum_belongings[count_bel-1]!=count){
 return LIP_ERROR;
}

k=count_bel-2;

while (k>=start){
   
   left=lefts[k];   
   right=rights[k];

   while(left<right){
       i=left;
       j=right;
       xh=bucket[sum_belongings[k]-1].h_index;

       do{
          while (bucket[i].h_index<xh ){i++;}
          while (bucket[j].h_index>xh){j--;}
          if (i<=j){
                    pomh=bucket[i];
                    bucket[i]=bucket[j];
                    bucket[j]=pomh;
	            i++;
        	    j--;
	  }
	  
       } while(i<=j);
       l=k;
       while (j<sum_belongings[l]-1) {lefts[l]=i;l--; if (l<0) {break;}}
       l=0; 
       while (i>sum_belongings[l]-1) {rights[l]=j;l++; if (l>k) {break;}} 
       left=lefts[k];
       right=rights[k];          
   }
   k--;
 }       
 return LIP_SUCCESS;    
}  
/*correct phase of in-core partitioner*/
int correct(int left_sum_sizes,
            LIP_bucket_element* bucket,
            int my_current_size,
	    int correct_size,
	    LIP_IDatamap  * irr_data){
int i,j,k,begin;
int begin_ownerN;
int end_owner0;
int *corr_owners=NULL,*belongings=NULL,*indices_to_change=NULL;
int count,new_size,new_count;
int owner0,ownerN;
int count_to_change,ind;
int *sizes=NULL,*displacements=NULL,*isizes=NULL,
   *idisplacements=NULL,*tr_table_owners=NULL;
LIP_ttable_change *ttab_owners=NULL,*change_to_ttab_owners=NULL;
int my_right_size;
int *trans_table;
MPI_Datatype change_type;
char* pname="correct";
if(1==LIP_env.size){
    return LIP_SUCCESS;
}

trans_table=irr_data->procs;
 /*who should own the first element in the bucket ?*/
 owner0=(int)((left_sum_sizes)/correct_size);
 /*where is the last element that should belong to owner0 ?*/
 end_owner0=owner0*correct_size+correct_size-1;
 /*who should own the last element in the bucket ?*/
 ownerN=(int)((my_current_size+left_sum_sizes-1)/correct_size);
 /*where is the first element that should belong to ownerN ?*/
 begin_ownerN=ownerN*correct_size;
 corr_owners=malloc (sizeof(int)*(ownerN-owner0+1));

 /*who is the correct owner of the data?*/
 LIP_Assert(corr_owners,pname,"during correction phase",LIP_OUT_OF_MEMORY);
 /*how many data should belong to the owner?*/
 belongings=malloc (sizeof(int)*(ownerN-owner0+1));
 LIP_Assert(belongings,pname,"during correction phase",LIP_OUT_OF_MEMORY);

 
 /*only one owner*/
 if (ownerN==owner0) {
  corr_owners[0]=owner0;
  belongings[0]=my_current_size;
 }
 else{
  corr_owners[0]=owner0;
  belongings[0]=end_owner0+1-left_sum_sizes;
  corr_owners[ownerN-owner0]=ownerN;
  belongings[ownerN-owner0]=my_current_size+left_sum_sizes-begin_ownerN;
  for(i=owner0+1 ;i<ownerN;i++){
   corr_owners[i-owner0]=i;
   belongings[i-owner0]=correct_size;
  }
 }
 
#ifdef DEBUG   
 for(i=0;i<(ownerN-owner0+1);i++){
  fprintf(debf,"dla nodu %d %d element(-y,-ow)\n",corr_owners[i],belongings[i]);
 }
#endif

/*count_to_change- how many entries in the translation table */
/*should be change (change the correct owner from our node to others)*/
 count_to_change=my_current_size;
 if (corr_owners[0]==LIP_env.rank){
     count_to_change-=belongings[0];
 }

 for(i=1;i<ownerN-owner0+1;i++){
  if(corr_owners[i]==LIP_env.rank){
   count_to_change-=belongings[i];
  }
  
/*Summing up the belongings.  After that operation*/
/*the belongings array will store positions in the*/
/*arrays hilbert_indices and indices that after sorting*/
/* will determine borders for new owners*/

  belongings[i]+=belongings[i-1];
 }
 
/*sorting the table partially */
 part_sort (belongings,ownerN-owner0+1,bucket,my_current_size);

#ifdef DEBUG   
 fprintf(debf,"Posortowane indeksy hilberta\n");
 for(i=0;i<my_current_size;i++){
  fprintf(debf,"%u ",bucket[i].h_index);
 }
 fprintf(debf,"\n");
 fprintf(debf,"odpowiadajace im indeksy globalne\n");
 for(i=0;i<my_current_size;i++){
  fprintf(debf,"%d ",bucket[i].index);
 }
 fprintf(debf,"\n");
#endif
 
 
/*determine who has entries in tha translation table*/
/*that should be changed*/

 tr_table_owners=malloc(count_to_change*sizeof(int));
 if(0==count_to_change){
  tr_table_owners=NULL;
 }
 else{
   LIP_Assert(tr_table_owners,pname,"during correction phase",LIP_OUT_OF_MEMORY);
 }
/*creation displacements and sizes for sending  the*/
/*data that should be changed to translation table owners*/
 
 begin=0;
 k=0;
 
 sizes=malloc(LIP_env.size*sizeof(int));
 LIP_Assert(sizes,pname,"during correction phase",LIP_OUT_OF_MEMORY);
 
 for(i=0;i<LIP_env.size;i++){
  sizes[i]=0;
 }
 displacements=malloc(LIP_env.size*sizeof(int));
 LIP_Assert(displacements,pname,"during correction phase",LIP_OUT_OF_MEMORY);

/*counting the owners of translation table */
 for(j=0;j<ownerN-owner0+1;j++){
  if (corr_owners[j]!=LIP_env.rank){
#ifdef DEBUG
 fprintf(debf,"indeksy globalne,ktorych dane zmieniaja wlasciciela z nodu naszego do nodu %d \n",corr_owners[j]);
#endif
   for(i=begin;i<belongings[j];i++){
    tr_table_owners[k]=DIV(bucket[i].index,correct_size);
    sizes[tr_table_owners[k]]++;
#ifdef DEBUG
    fprintf(debf,"%d jest u:",bucket[i].index);
    fprintf(debf,"%d\n",tr_table_owners[k]);
#endif

    k++;
   }
  }
#ifdef DEBUG
 fprintf(debf,"\n");
#endif
  begin=belongings[j];
 }

 displacements[0]=0;
 sizes[LIP_env.rank]=0;

 for(i=1;i<LIP_env.size;i++){
  displacements[i]=displacements[i-1]+sizes[i-1];
  sizes[i-1]=0;
 }
 
 
/*create tables for sending the correct owners and global indices*/
/*to the owners of the translation table*/
/*in order to let them change the entries*/
 new_count=displacements[LIP_env.size-1]+sizes[LIP_env.size-1];
 change_to_ttab_owners=malloc(new_count*sizeof(LIP_ttable_change));
 if(new_count){
  LIP_Assert(change_to_ttab_owners,pname,"during correction phase",LIP_OUT_OF_MEMORY);
 }

 sizes[LIP_env.size-1]=0;

 begin=0;
 k=0;
 for(j=0;j<ownerN-owner0+1;j++){
  if (corr_owners[j]!=LIP_env.rank){
   for(i=begin;i<belongings[j];i++){
   /*we are not the translation table owner -*/
   /* we need to send the change to other node*/
    if (tr_table_owners[k]!=LIP_env.rank){
     change_to_ttab_owners[displacements[tr_table_owners[k]]+
                           sizes[tr_table_owners[k]]].g_index=bucket[i].index;
     change_to_ttab_owners[displacements[tr_table_owners[k]]+
                           sizes[tr_table_owners[k]]].owner=corr_owners[j];
     sizes[tr_table_owners[k]]++;
    }
    else{
    /*we change translation table on-line*/

     ind=REM(bucket[i].index,correct_size);
     trans_table[ind]=corr_owners[j];
#ifdef DEBUG
fprintf(debf,"zmieniam on-line trans table ind global %d local %d na ownera %d\n",
              bucket[i].index,ind,corr_owners[j]);
#endif 
    }
    k++;
   }
   }
  begin=belongings[j];
 }
#ifdef DEBUG
fprintf(debf,"zmieniona on-line tablica translacji\n");

for(i=0;i<irr_data->ttable_ic_size;i++){
 fprintf(debf,"%d ",trans_table[i]);
}
fprintf(debf,"\n");

fprintf(debf,"do wyslania nowi wlasciciele\n");
for(j=0;j<LIP_env.size;j++){
 fprintf(debf,"do nodu %d\n",j);
 for(i=displacements[j];i<displacements[j]+sizes[j];i++){
  fprintf(debf,"global index %d new owner %d \n",change_to_ttab_owners[i].g_index,
  change_to_ttab_owners[i].owner);
 }
}
 fprintf(debf,"\n");


#endif
 
/*communication*/
 isizes=malloc(sizeof(int)*LIP_env.size);
 LIP_Assert(isizes,pname,"during correction phase",LIP_OUT_OF_MEMORY);

 idisplacements=malloc(sizeof(int)*LIP_env.size);
 LIP_Assert(idisplacements,pname,"during correction phase",LIP_OUT_OF_MEMORY);

 MPI_Alltoall(sizes,1,MPI_INT,isizes,1,MPI_INT,LIP_env.comm);
 
 idisplacements[0]=0;
 for(i=1;i<LIP_env.size;i++){
 
  idisplacements[i]=idisplacements[i-1]+isizes[i-1];
 }
 new_size=idisplacements[LIP_env.size-1]+isizes[LIP_env.size-1];
 free(corr_owners);
 free(indices_to_change);
 
 ttab_owners=malloc(sizeof(LIP_ttable_change)*(new_size));
 
 if(new_size!=0){
 
  LIP_Assert(ttab_owners,pname,"during correction phase",LIP_OUT_OF_MEMORY);
 
 }
 
 MPI_Type_contiguous(2,MPI_INT,&change_type);
 
 MPI_Type_commit(&change_type); 		     		     
 
 MPI_Alltoallv(change_to_ttab_owners,sizes,displacements,
               change_type,ttab_owners,isizes,idisplacements,
	       change_type,LIP_env.comm);
 
 

/*changing the translation table */
 for(i=0;i<new_size;i++){
    
  trans_table[REM(ttab_owners[i].g_index,correct_size)]=
 
  ttab_owners[i].owner;
  
 
 
 }
#ifdef DEBUG

fprintf(debf,"do zmiany przez nas nowi wlasciciele\n");
for(i=0;i<idisplacements[LIP_env.size-1]+isizes[LIP_env.size-1];i++){
 fprintf(debf,"%d ",ttab_owners[i].owner);
}
 fprintf(debf,"\n");

fprintf(debf,"indeksy globalne, ktorych dane zmieniaja wlascicieli\n");
for(i=0;i<idisplacements[LIP_env.size-1]+isizes[LIP_env.size-1];i++){
 fprintf(debf,"%d ",ttab_owners[i].g_index);
}
fprintf(debf,"\n");
fprintf(debf,"zmieniona tablica translacji\n");
for(i=0;i<irr_data->ttable_ic_size;i++){
 fprintf(debf,"%d ",trans_table[i]);
}
fprintf(debf,"\n");

#endif

 if (isizes!=NULL) {free(isizes);}
 if (idisplacements!=NULL){free(idisplacements);}
 if (sizes!=NULL) free(sizes);
 if (idisplacements!=NULL) free(displacements);

 if (change_to_ttab_owners!=NULL) free(change_to_ttab_owners);	       
 if (ttab_owners!=NULL) free(ttab_owners);	       
 return LIP_SUCCESS;
}

