/*LIP library 
hilbert.c - contains functions for counting hilbert index
used by partitioner
created by Katarzyna Zajac 
*/

#include <stdio.h>
#include <math.h>
#include "hilbert.h"

#define SCHEMES_2D 8
#define SCHEMES_3D 24
#define SQUARES 4
#define CUBES 8

/*counting hilbert index in 2 dim space*/
void hilbert2d(int count,
double* pointX,double* pointY,double vertexX,double vertexY,  
double edge, double minEdge,unsigned long* hilbert_index)
{
 

  int square;
  int i,k,j,nr_i;
  double * x=pointX;
  double * y=pointY;
  int h_prec=sizeof(unsigned long)*8-2;
  unsigned long index_h=0;
  int my_scheme=0;
  double curr_verX,curr_verY,curr_edge;
  
  unsigned long index_add[SCHEMES_2D][SQUARES]=
  {
   {0,3,1,2},
   {3,0,2,1},
   {0,1,3,2},
   {1,0,3,2},
   {2,3,1,0},
   {3,2,0,1},
   {1,2,0,3},
   {2,1,3,0}
  };
  short scheme[SCHEMES_2D][SQUARES]=
  {
   {2,4,0,0},
   {5,3,1,1},
   {0,2,7,2},
   {3,1,3,6},
   {4,0,4,7},
   {1,5,6,5},
   {6,6,5,3},
   {7,7,2,4}
  };
  for(j=0;j<SCHEMES_2D;j++)
    for(k=0;k<SQUARES;k++)
     index_add[j][k]=(index_add[j][k])<< h_prec;
  nr_i=0;
  for(i=0;i<count;i++){
   index_h=0;
   my_scheme=0;
   curr_verX=vertexX;
   curr_verY=vertexY;
   curr_edge=edge;
   
   for(j=0;j<SCHEMES_2D;j++)
    for(k=0;k<SQUARES;k++)
     index_add[j][k]=(index_add[j][k])<< (2*nr_i);

     nr_i=0; 
     while (curr_edge>minEdge){
     nr_i++;
     curr_edge/=2.0; 
     square=(x[i] > (curr_verX + curr_edge))+
            2*(y[i] > (curr_verY + curr_edge));
     
     index_h = index_h+index_add[my_scheme][square];
     my_scheme=scheme[my_scheme][square];
     
 
     curr_verX=curr_verX+(curr_edge* (square&0x1));
     curr_verY=curr_verY+(curr_edge* ((square>>1)&0x1));

     for(j=0;j<SCHEMES_2D;j++)
      for(k=0;k<SQUARES;k++)
       index_add[j][k]=index_add[j][k]>>2;
      
   }
   index_h=index_h >>(sizeof(long)*8-2*(nr_i));
   hilbert_index[i]=index_h;
  }
}

/*counting hilbert index in 3-dim space*/
void hilbert3d(int count,
double* pointX,double* pointY,double* pointZ,
double vertexX,double vertexY,double vertexZ,  
double edge, double minEdge,
unsigned long* hilbert_index)
{
  int cube;
  int i,k,j,nr_i;
  double * x=pointX;
  double * y=pointY;
  double * z=pointZ;
  int h_prec=sizeof(unsigned long)*8-3;
  unsigned long index_h=0;
  int my_scheme=0;
  double curr_verX,curr_verY,curr_verZ,curr_edge;
  
unsigned long index_add[SCHEMES_3D][CUBES]=
{
{3,0,2,1,4,7,5,6},
{4,7,5,6,3,0,2,1},
{7,4,6,5,0,3,1,2},
{0,3,1,2,7,4,6,5},
{2,3,5,4,1,0,6,7},
{5,4,2,3,6,7,1,0},
{3,2,4,5,0,1,7,6},
{4,5,3,2,7,6,0,1},
{1,0,6,7,2,3,5,4},
{6,7,1,0,5,4,2,3},
{0,1,7,6,3,2,4,5},
{7,6,0,1,4,5,3,2},
{5,2,4,3,6,1,7,0},
{2,5,3,4,1,6,0,7},
{1,6,0,7,2,5,3,4},
{6,1,7,0,5,2,4,3},
{5,6,4,7,2,1,3,0},
{2,1,3,0,5,6,4,7},
{6,5,7,4,1,2,0,3},
{1,2,0,3,6,5,7,4},
{7,0,6,1,4,3,5,2},
{0,7,1,6,3,4,2,5},
{3,4,2,5,0,7,1,6},
{4,3,5,2,7,0,6,1},
};
short scheme[SCHEMES_3D][CUBES]=
{
{18,9,14,14,18,4,23,23},
{19,8,15,15,19,5,22,22},
{11,17,20,20,6,17,22,22},
{10,16,21,21,7,16,23,23},
{1,8,16,8,1,23,16,13},
{0,9,17,9,0,22,17,12},
{8,2,8,16,22,2,12,16},
{9,3,9,17,23,3,13,17},
{0,20,17,14,0,6,17,6},
{1,21,16,15,1,7,16,7},
{21,3,15,18,4,3,4,18},
{20,2,14,19,5,2,5,19},
{19,17,15,20,19,17,6,5},
{18,16,4,21,18,16,6,4},
{19,17,11,8,19,17,22,22},
{18,16,10,9,18,16,23,23},
{14,14,2,8,12,12,2,5},
{15,15,3,9,13,13,3,4},
{15,15,10,1,13,13,7,1},
{14,14,11,0,12,12,6,0},
{11,8,2,0,12,12,2,0},
{10,9,3,1,13,13,3,1},
{14,14,2,0,6,5,2,0},
{15,15,3,1,7,4,3,1},
};
  for(j=0;j<SCHEMES_3D;j++)
    for(k=0;k<CUBES;k++)
     index_add[j][k]=(index_add[j][k])<< h_prec;
  nr_i=0;
  for(i=0;i<count;i++){
   index_h=0;
   my_scheme=10;
   curr_verX=vertexX;
   curr_verY=vertexY;
   curr_verZ=vertexZ;
   curr_edge=edge;
   
   for(j=0;j<SCHEMES_3D;j++)
    for(k=0;k<CUBES;k++)
     index_add[j][k]=(index_add[j][k])<< (3*nr_i);

     nr_i=0; 
     while (curr_edge>minEdge){
     nr_i++;
     curr_edge/=2.0; 
     cube=(x[i]>(curr_verX + curr_edge))+
            2*(y[i] > (curr_verY + curr_edge))+
	    4*(z[i] >(curr_verZ+curr_edge));
     
     index_h = index_h+index_add[my_scheme][cube];
     my_scheme=scheme[my_scheme][cube];
     
 
     curr_verX=curr_verX+(curr_edge* (cube&0x1));
     curr_verY=curr_verY+(curr_edge* ((cube>>1)&0x1));
     curr_verZ=curr_verZ+(curr_edge* ((cube>>2)&0x1));

     for(j=0;j<SCHEMES_3D;j++)
      for(k=0;k<CUBES;k++)
       index_add[j][k]=index_add[j][k]>>3;
      
   }
   index_h=index_h >>(sizeof(long)*8-3*(nr_i));
   hilbert_index[i]=index_h;
 }

}

