/* 
LIP Library
File ooc.c 
created by Maciej Malawski 
5.04.2000
*/

#include <stdlib.h>
#include <lipimpl.h>
#include <liptools.h>
#include <debug.h>
/****************************************************************************
Function LIP_OOC_localize transforms local indices from ilocal array into 
array itlocal. It uses and modifies LIP_IOBufmap structure, which maps
transformed local indices (that point to data in memory) to local indices 
corresponding to data on a disk.
Count specifies size of memory buffer (size of ilocal and itlocal arrays.
Function returns LIP_OUT_OF_MEMORY if it cannot allocate enough memory 
buffers, LIP_OK if everything went OK and LIP_ERROR if other errors happened. 
*****************************************************************************/

int
LIP_OOC_localize(void *ilocal, LIP_Indextype iltype, void *itlocal,
                 LIP_Indextype ittype, int count, int flags, void *buf, 
					  
		 MPI_Datatype datatype, LIP_IOBufmap *bmap, 
		 LIP_Schedule *schedule)
{
  LIP_IIOBufmap * bm;
  LIP_ISchedule * s;
  int * tmp;
  int sumka=0,sumkap=0;
  int i, j, next;
  char * oldnew;
  int * l_indcs, * tl_indcs;
  int newhash_index, oldhash_index, ind;
  int empty;
  int first_free;
  int elocations_size;
  int ret;
  char* pname="LIP_OOC_localize";
  /*if(LIP_env.size==1){
   LIP_Error_ret(pname,"for a single node",LIP_NOT_SUPPORTED);
   return LIP_NOT_SUPPORTED;
  } */
  if(count<0){
   LIP_Error_ret(pname,"count<0!!!",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }  
  if ((LIP_INDEXTYPE_INT!=iltype)||(LIP_INDEXTYPE_INT!=ittype)){
   LIP_Error_ret(pname,"indextype",LIP_NOT_SUPPORTED);
   return LIP_NOT_SUPPORTED;
  }
  l_indcs = (int *) ilocal;
  tl_indcs = (int *) itlocal;
  
  /* Get pointers to IO_Bufmap and Schedule structures */  
  bm = LIP_IOBufmap_convert(*bmap);
  if (NULL==bm){
   LIP_Error_ret(pname,"invalid IOBufmap handle",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  
  s = LIP_Schedule_convert(*schedule);
  if(NULL==s){
   LIP_Error_ret(pname,"invalid schedule handle",LIP_INVALID_ARGUMENT);
   return LIP_INVALID_ARGUMENT;
  }
  /* If tables were not allocated we have to do it now. */
  if (NULL == bm->oldhash_val)
  {
    bm->htab_size = prime_round_up( (bm->size) * 1.5 );
    
    if(NULL==(bm->oldhash_val = 
    malloc( (sizeof *(bm->oldhash_val)) * bm->htab_size ))){
     LIP_Error_ret(pname,"too big size of memory data buffer",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    
    if(NULL==(bm->oldhash_key = malloc( (sizeof *(bm->oldhash_key)) * bm->htab_size ))){
     LIP_Error_ret(pname,"too big size of memory data buffer",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    
  }
  if (NULL == bm->newhash_val)
  {
    bm->newhash_val = malloc( (sizeof *(bm->newhash_val)) * bm->htab_size );
    if (NULL == bm->newhash_val){
     LIP_Error_ret(pname,"too big size of memory data buffer",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    bm->newhash_key = malloc( (sizeof *(bm->newhash_val)) * bm->htab_size );
    if (NULL == bm->newhash_key){
     LIP_Error_ret(pname,"too big size of memory data buffer",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY;
    }
    for (i = 0; i < bm->htab_size; i++)
    {
      bm->newhash_val[i] = HASH_EMPTY; /* a special value which means
                                   that position is empty */
      bm->newhash_key[i] = -1;
    }
  }
  
  /* Swap old and new hash tables*/
  tmp = bm->newhash_val;
  bm->newhash_val = bm->oldhash_val;
  bm->oldhash_val = tmp;
  tmp = bm->newhash_key;
  bm->newhash_key = bm->oldhash_key;
  bm->oldhash_key = tmp;
  
  /* New hashtable has to be cleared */
  for (i = 0; i < bm->htab_size; i++) /* clear new hash*/
  {
    bm->newhash_val[i] = HASH_EMPTY; /* a special value which means
                                   that position is empty */
    bm->newhash_key[i] = -1;                                   
  }
  
  /* 2. Now we have to find what data we need in next step of calculations */

  /* 2.1. Create oldnew array and fill it with zeroes.
     Oldnew[i] == 1  means data indexed by i is needed for 
     further calculations and should be left. Zero means data
     is not needed and place can be freed.
  */
  oldnew = malloc ((sizeof (*oldnew)) * bm->size);
  if (NULL == oldnew) {
   LIP_Error_ret(pname,"too big size of memory data buffer",LIP_OUT_OF_MEMORY);
   return LIP_OUT_OF_MEMORY;
  }
  for (i = 0; i < bm->size; i++) oldnew[i] = 0;
  
  /* 2.3 Here we have to fill oldnew array with indices 
     from ilocal array (named l_indcs)
     We skip indices pointing to ghost area.
     If we find index that has not been placed in new hash table
     we check if it existed in old hash table.
        If it was not in old we reserva a new place in new.
        If it was, we copy its location from old and mark it in oldnew
        with 1.
  */
  
  for (i=0; i < count; i++)  /* for every element in ilocal[] */ 
  {
    if(l_indcs[i] < s->data_size) /* skip ghost area */
    {
      sumkap++;
      if (LIP_SUCCESS!= hash_query (l_indcs[i], &newhash_index, &empty, bm->newhash_val, 
                        bm->newhash_key, bm->htab_size)){
         LIP_Error_ret(pname,"error",LIP_ERROR);
         return LIP_ERROR;
      }                  
      if (empty)
      {
        sumka++;
        if(LIP_SUCCESS != hash_query (l_indcs[i], &oldhash_index, &empty, bm->oldhash_val, 
                    bm->oldhash_key, bm->htab_size)){
          LIP_Error_ret(pname,"error",LIP_ERROR);
          return LIP_ERROR;
	}		  
        if (empty)
        {
          bm->newhash_val[newhash_index] = HASH_RESERVED;
          bm->newhash_key[newhash_index] = l_indcs[i];
        }
        else
        {
          ind = bm->newhash_val[newhash_index] = bm->oldhash_val[oldhash_index];
          bm->newhash_key[newhash_index] = l_indcs[i];
          oldnew[ind] = 1;
        }
      }		
    } /* end if */  
  } /* end for every element in ilocal[]*/
  
  /* 2.4 We do with schedule->elocations the same as in 2.3 with ilocal[] */
  
  elocations_size = s->esizes[LIP_env.size - 1] + s->edispls[LIP_env.size - 1];
  for (i=0; i < elocations_size; i++)  /* for every element in s->elocations[] */ 
  {
    if(LIP_SUCCESS!= hash_query (s->elocations[i], &newhash_index, &empty, 
                      bm->newhash_val, bm->newhash_key, bm->htab_size)){
      LIP_Error_ret(pname,"error",LIP_ERROR);
      return LIP_ERROR;
    }
    if (empty)
    {
      if(LIP_SUCCESS!= hash_query (s->elocations[i], &oldhash_index, &empty, 
                        bm->oldhash_val, bm->oldhash_key, bm->htab_size)){		  
          LIP_Error_ret(pname,"error",LIP_ERROR);
          return LIP_ERROR;
      }
      if (empty)
      {
        bm->newhash_val[newhash_index] = HASH_RESERVED;
        bm->newhash_key[newhash_index] = s->elocations[i];
      }
      else
		  {
        ind = bm->newhash_val[newhash_index] = bm->oldhash_val[oldhash_index];
        bm->newhash_key[newhash_index] = s->elocations[i];
        oldnew[ind] = 1;
      }
		}
  } /* end for every element in schedule->elocations[]*/  
  
  /* 3. Now we know which places are to be used and which are free.
        We have to update tables - put data in reserved places */

  /* find first free place (0) in oldnew[] */
  j=0;
  while ((j< bm->size) && oldnew[j]) j++; 
  if (bm->size == j) {
     LIP_Error_ret(pname,"too small memory buffer",LIP_OUT_OF_MEMORY);
     return LIP_OUT_OF_MEMORY; /* FIXME error exit? */
  }
  else first_free = j;
  /* 3.1. Updating indexes from ilocal.
        If index is not from ghost area and we reserved place for it in hash table
          we have to find free place in memory buffer to assign it to that index.
          We use for it first free index as marked in oldnew array
          If index already had its position in memory, 
             we put it straight into tlocal array
        If index points to ghost area (what means "disk ghost area"
          then we move it to the left to fit into memory ghost area. */
  next = 0;
  for (i=0; i < count; i++)            /* for every element in ilocal[] */ 
  {
    if(l_indcs[i] < s->data_size)      /* if not in ghost area */
    {
      if(LIP_SUCCESS!=hash_query (l_indcs[i], &newhash_index, &empty, bm->newhash_val, 
                  bm->newhash_key, bm->htab_size)){
        
	LIP_Error_ret(pname,"error",LIP_ERROR);
     
	return LIP_ERROR;
      }
      if (HASH_RESERVED == bm->newhash_val[newhash_index])
      {
        if (bm->size == first_free) {
                LIP_Error_ret(pname,"too small memory buffer for data needed by this node",LIP_OUT_OF_MEMORY);
                return LIP_OUT_OF_MEMORY; /* FIXME error exit? */
        }
        tl_indcs[i] = bm->newhash_val[newhash_index] = first_free;
        /* find next "0" in oldnew */
        j=first_free+1;
        while ((j< bm->size) && oldnew[j] ) j++;
        first_free = j;
      }
      else tl_indcs[i] = bm->newhash_val[newhash_index];

    } /* end if */
    else                               /* if in ghost area */
      tl_indcs[i] = l_indcs[i] - (s->data_size - bm->size);  
  } /* end for every element in ilocal[]*/  

  /* 3.2. Here we do the same with indices from schedule->elocations as
          we have done with ilocal inices in step 3.1
  */
  for (i=0; i < elocations_size; i++)  /* for every element in s->elocations[] */ 
  {
    if (LIP_SUCCESS!= hash_query (s->elocations[i], &newhash_index, &empty, bm->newhash_val, 
                bm->newhash_key, bm->htab_size)){
	LIP_Error_ret(pname,"error",LIP_ERROR);
     	return LIP_ERROR;
    }
    if (HASH_RESERVED == bm->newhash_val[newhash_index])
    {
      if (bm->size == first_free) {

         LIP_Error_ret(pname,"too small memory buffer for data needed by other nodes from this node",LIP_OUT_OF_MEMORY);
         return LIP_OUT_OF_MEMORY; /* FIXME error exit? */
      } 
      s->elocations[i] = bm->newhash_val[newhash_index] = first_free;

      /* find next "0" in oldnew */
      j=first_free+1;
      while ((j< bm->size) && oldnew[j]) j++;
      first_free = j;

    }
    else s->elocations[i] = bm->newhash_val[newhash_index];
  } /* end for every element in schedule->elocations */

  free (oldnew);
  return LIP_SUCCESS;
} /* end LIP_OOC_localize */ 
  


  
