/*LIP library
partincore.c -contains in-core partitioner
creted by Katarzyna Zajac
*/
#include  <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include <lip.h>
#include <lipimpl.h>
#include <correct.h>
#include <debug.h>
#include <sys/times.h>
/*main function that creates distribution*/
int LIP_create_hilbert_distribution_in_core
( 
/*current data distribution*/
  LIP_Datamap datamap,

/*new data distribution*/
  LIP_Datamap *new_datamap,
  
/*the coordinates of the points */
  double *coX,double *coY,double *coZ,

/*data for counting hilbert indices*/
  double* vertex,double edge,double minedge,

/*quantity of the sample*/
  int sample_count
)
{

  MPI_Datatype element_type,types[2]={MPI_INT,MPI_UNSIGNED};

  int disp[2],block[2]={1,1};
  
  LIP_bucket_element *my_elements,*buckets;
  
  int count,ret;

  int mem_size=10;

  int my_offset,my_size,my_size_all,j,corr_size;

  int l,n,npol;

  int *sizes,*isizes,*idisplacements,*displacements,*index_buckets,*my_index,
       *corr_proc,*procs;

  unsigned long *h_index,*h_index_all,*borders,*h_index_buckets,
                *h_index_work,*my_h_index;

  double *samplex,*sampley,*samplez,*coX_work,*coY_work,*coZ_work;

  int sample_index;

  MPI_File fx,fy;
  
  MPI_Datatype index_type=MPI_DOUBLE;
  
  MPI_Datatype data_type=MPI_DOUBLE;
  
  MPI_Status status;
  
  double* ver;

 
  double vertexX,vertexY,vertexZ,rr;
  
int ilosc,left_sum,pass;     
  
  LIP_IDatamap* dm,*in_dm;
  
  MPI_Datatype sample_type,send_type;
  
  int i;
  
  /*char*ghost_area_r=getenv("GHOST_AREA_RATE");*/
  /*int GHOST_AREA_RATE;  */
  char*pname="LIP_create_hilbert_distribution_in_core";

  int rank=LIP_env.rank;
  
  /*getting information about current */
  /*distribution (where should  we start to read?)*/

  if(LIP_env.size==1){
   return LIP_SUCCESS;
  }  
  /*if (NULL==ghost_area_r){
    GHOST_AREA_RATE=2;
  }
  else{
    GHOST_AREA_RATE=atoi(ghost_area_r);
  }*/
  ver=(double*)vertex;
  /*printf("%d\n",GHOST_AREA_RATE); */
  /*creating MPI datamap for sending the bucket element*/
  
  disp[0]=0;
  
  disp[1]=sizeof(int);
  
  
  MPI_Type_struct(2,block,disp,types,&element_type);
  
  vertexX=ver[0];
  
  vertexY=ver[1];
  
  if (coZ!=NULL){vertexZ=ver[2];}
  
  dm=LIP_Datamap_convert(datamap);
  
  LIP_Assert(dm,pname,"old datamap",LIP_INVALID_ARGUMENT);
  
  if (dm->type!=LIP_DATAMAP_BLOCK){
  
   LIP_Error_quit(pname,"the new distribution can be created only for BLOCK datamap",LIP_NOT_SUPPORTED);
  
   return LIP_NOT_SUPPORTED;

  }
  
  count=dm->node_size;
  
  ret=LIP_Datamap_create(LIP_DATAMAP_INDIRECT,dm->global_size,&(dm->local_size),
                     LIP_INDEXTYPE_INT,1,LIP_NONLOCAL,new_datamap);
  
  if(ret!=LIP_SUCCESS){
  
   LIP_Error_quit(pname,NULL,ret);
 
  }
  
  in_dm=LIP_Datamap_convert(*new_datamap);
  
  in_dm->ttable_ic_size=dm->node_size;
  
  in_dm->ttable_ic_real_size=dm->node_size*GHOST_AREA_RATE;
  
  in_dm->procs=malloc(in_dm->ttable_ic_real_size*sizeof(int));
  
  LIP_Assert(in_dm->procs,pname,"not enough place for translation table-try to decrease GHOST_AREA_RATE",LIP_OUT_OF_MEMORY);
              
  
  procs=in_dm->procs;
 
  my_offset=(dm->local_size) * LIP_env.rank;
    
  /*choose sample*/
  if(0>=sample_count){
   LIP_Error_quit(pname,"sample count",LIP_INVALID_ARGUMENT);
  }
  h_index=(unsigned long*)(malloc(sizeof(unsigned long)*sample_count));
 
  LIP_Assert (h_index,pname,NULL,LIP_OUT_OF_MEMORY);
   
  samplex=(double*)(malloc(sizeof(double)*sample_count));
  
  LIP_Assert (samplex,pname,NULL,LIP_OUT_OF_MEMORY);
 
  sampley=(double*)(malloc(sizeof(double)*sample_count));
  
  LIP_Assert (sampley,pname,NULL,LIP_OUT_OF_MEMORY);
 
  if(coZ!=NULL){ 
    samplez=(double*)(malloc(sizeof(double)*sample_count));
    
    LIP_Assert (samplez,pname,NULL,LIP_OUT_OF_MEMORY);
 
  }
  
  srand(5*LIP_env.rank);
  
  if(coZ!=NULL){
  
   for(i=0;i<sample_count;i++){
    rr=rand();
#ifdef DEBUG    
    fprintf(debf,"%d ",rr);
#endif
    sample_index=(int)(count*rr/(RAND_MAX+1.0));

    samplex[i]=coX[sample_index];

    sampley[i]=coY[sample_index];

    samplez[i]=coZ[sample_index];
   }

  } 
  else{
   for(i=0;i<sample_count;i++){

    rr=rand();
#ifdef DEBUG    
    fprintf(debf,"%d ",rr);
#endif

    sample_index=(int)(count*rr/(RAND_MAX+1.0));

    samplex[i]=coX[sample_index];

    sampley[i]=coY[sample_index];

   }
  }

#ifdef DEBUG
  fprintf(debf,"Wspolrzedne\n");
  for(i=0;i<sample_count;i++){
    fprintf(debf,"%1.1f ",samplex[i]);
  } 
  fprintf(debf,"\n"); 
  
  
  for(i=0;i<sample_count;i++){
    fprintf(debf,"%1.1f ",sampley[i]);
  } 
  fprintf(debf,"\n"); 
  if(coZ!=NULL){
   for(i=0;i<sample_count;i++){
    fprintf(debf,"%1.1f ",samplez[i]);
   }
   
   fprintf(debf,"\n"); 
  }
#endif  
  /*count hilbert index*/
  if (coZ!=NULL){

   hilbert3d(sample_count,samplex,sampley,samplez,vertexX,vertexY,vertexZ,edge,minedge,
                  h_index);

   free(samplez);
  }
  else{

   hilbert2d(sample_count,samplex,sampley,vertexX,vertexY,edge,minedge,
                    h_index);

  }
  free(samplex);

  free(sampley);

#ifdef DEBUG  
  fprintf(debf,"Indeksy Hilberta:\n");
  for(i=0;i<sample_count;i++){
    fprintf(debf,"%u ",h_index[i]);
  } 
  fprintf(debf,"\n"); 
#endif  
  /*sorting hilbert indices */

  qsort(h_index,sample_count,sizeof(unsigned long),
  (int (*)(const void *,const void *))sortLD);
 
#ifdef DEBUG 
   fprintf(debf,"Posortowane indeksy Hilberta\n");
   for(i=0;i<sample_count;i++){
    fprintf (debf,"%u ",h_index[i]);
   }
   fprintf(debf,"\n");
#endif
  
  /*merge indices*/
  
  h_index_all=chain_merge(h_index,sample_count,MPI_UNSIGNED_LONG);
  
   
  if(LIP_env.rank==0){
#ifdef DEBUG
   fprintf(debf,"Posortowane indeksy Hilberta po scaleniu\n");
  
   for(i=0;i<sample_count*LIP_env.size;i++){
    fprintf (debf,"%u ",h_index_all[i]);
   }
   fprintf(debf,"\n");
#endif

 /*count borders*/

   borders=malloc(sizeof(unsigned long)*(LIP_env.size));
   
   LIP_Assert (borders,pname,NULL,LIP_OUT_OF_MEMORY);
   
   borders[0]=0;
   
   for(i=1;i<LIP_env.size;i++){

    borders[i]=h_index_all[i*sample_count];

   }
   
#ifdef DEBUG   
   for(i=0;i<LIP_env.size;i++){
      fprintf (debf,"%u ",borders[i]);
   }
   fprintf(debf,"\n");
#endif

  }
  else{
     borders=malloc(sizeof(unsigned long)*(LIP_env.size));
 
     LIP_Assert (borders,pname,NULL,LIP_OUT_OF_MEMORY);

  }
  free (h_index_all);

  MPI_Bcast(borders,LIP_env.size,MPI_UNSIGNED_LONG,0,LIP_env.comm);
#ifdef DEBUG
  fprintf(debf,"granice:\n");
  for(i=0;i<LIP_env.size;i++){
      fprintf (debf,"%u ",borders[i]);
  }
  fprintf(debf,"\n");
#endif  

/*putting all points into buckets*/
/*counting hilbert indices and determining their process numbers*/
 
 coX_work=coX;
 coY_work=coY;
 h_index=realloc(h_index,sizeof(unsigned long)*count);
 
 LIP_Assert (h_index,pname,NULL,LIP_OUT_OF_MEMORY);

 /*h_index_buckets=malloc(sizeof(unsigned long)*count);*/
 /*index_buckets=malloc(sizeof(unsigned long)*count);*/
 h_index_work=h_index;
 
 sizes=malloc(LIP_env.size*sizeof(int));
 
 LIP_Assert (sizes,pname,NULL,LIP_OUT_OF_MEMORY);

 isizes=malloc(LIP_env.size*sizeof(int));
 
 LIP_Assert (isizes,pname,NULL,LIP_OUT_OF_MEMORY);

 for(i=0;i<LIP_env.size;i++){
  sizes[i]=0;
  isizes[i]=0;
 }
 displacements=malloc(LIP_env.size*sizeof(int));
 
 LIP_Assert (displacements,pname,NULL,LIP_OUT_OF_MEMORY);

 idisplacements=malloc(LIP_env.size*sizeof(int));
 
 LIP_Assert (idisplacements,pname,NULL,LIP_OUT_OF_MEMORY);
 /*counting hilbert indices for the elements and determine their new owners*/
 if(coZ!=NULL){
  coZ_work=coZ;
  hilbert3d(count,coX_work,coY_work,coZ_work,vertexX,vertexY,vertexZ,edge,minedge,
                  h_index_work);
  for(i=0;i<count;i++){
   l=0;
   while (borders[l]<=(*h_index_work)){
    l++; 
    if (LIP_env.size==l) {break;}
   }
   procs[i]=l-1;
   sizes[procs[i]]++;
   coX_work++;
   coY_work++;
   coZ_work++;
   h_index_work++;
  }
 }
 else{
   hilbert2d(count,coX_work,coY_work,vertexX,vertexY,edge,minedge,
                  h_index_work);
   for(i=0;i<count;i++){

   l=0;
   while (borders[l]<=(*h_index_work)){
    l++; 
    if (LIP_env.size==l) {break;}
   }
   procs[i]=l-1;
   sizes[procs[i]]++;
   coX_work++;
   coY_work++;
   h_index_work++;
  }
 }
 free (borders);
#ifdef DEBUG
 fprintf(debf,"indeksy hilberta:\n");
  for(i=0;i<count;i++){
      fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");

 fprintf(debf,"naleza do procesorow:\n");
   for(i=0;i<count;i++){
      fprintf (debf,"%u ",procs[i]);
  } 
  fprintf(debf,"\n");

#endif
/*putting elements to the buckets*/
 my_size=sizes[LIP_env.rank];
 my_elements=malloc(my_size*sizeof(*my_elements));
 buckets=malloc(sizeof(*buckets)*(count-my_size));
 
 LIP_Assert (buckets,pname,NULL,LIP_OUT_OF_MEMORY);
 
 sizes[LIP_env.rank]=0;
  
 displacements[0]=0;
 for(i=1;i<LIP_env.size;i++){
   displacements[i]=displacements[i-1]+sizes[i-1];
   sizes[i-1]=0;
 }
 displacements[LIP_env.rank]=displacements[LIP_env.size-1]
                             +sizes[LIP_env.size-1];
 sizes[LIP_env.size-1]=0;
  
 j=0;
 for(i=0;i<count;i++){
  if(procs[i]==LIP_env.rank){
   my_elements[j].index=i+my_offset;
   my_elements[j++].h_index=h_index[i];
  }
  else{ 
   buckets[displacements[procs[i]]+sizes[procs[i]]].h_index=h_index[i];
   buckets[displacements[procs[i]]+sizes[procs[i]]].index=i+my_offset;
   sizes[procs[i]]++;
  }
 }
#ifdef DEBUG
 fprintf(debf,"indeksy hilberta w kubelkach:\n");
  for(i=0;i<count-my_size;i++){
      fprintf (debf,"%u ",buckets[i].h_index);
  }
  fprintf(debf,"\n");
#endif
 
 /*communication-sending buckets to their owners*/
 
 MPI_Alltoall(sizes,1,MPI_INT,isizes,1,MPI_INT,LIP_env.comm);
 idisplacements[0]=0;
 isizes[LIP_env.rank]=0;
 for(i=1;i<LIP_env.size;i++){
  idisplacements[i]=idisplacements[i-1]+isizes[i-1];
 }

 my_size_all=my_size+idisplacements[LIP_env.size-1]+isizes[LIP_env.size-1];

 my_elements=realloc(my_elements,my_size_all*sizeof(*my_elements));

 LIP_Assert (my_elements,pname,NULL,LIP_OUT_OF_MEMORY);

 MPI_Alltoallv(buckets,sizes,displacements,element_type,
 my_elements+my_size,
 isizes,idisplacements,element_type,LIP_env.comm);


 free(buckets);
 

 free(isizes);

 free(sizes);

 free(displacements);

 free(idisplacements); 
#ifdef DEBUG
 fprintf(debf,"moje indeksy hilberta po otrzymaniu od innych:\n");
  for(i=0;i<my_size_all;i++){
      fprintf (debf,"%u ",my_elements[i].h_index);
  }
  fprintf(debf,"\n");
#endif
  
 MPI_Scan(&my_size_all,&left_sum,1,MPI_INT,MPI_SUM,LIP_env.comm);
 
 left_sum-=my_size_all;
 
 corr_size=dm->local_size;
 /*correction phase - see corect.c*/
 ret=correct(left_sum,my_elements,my_size_all,corr_size,in_dm);

 if(LIP_SUCCESS!=ret){
  LIP_Error_ret(pname,NULL,ret);
  return ret;

 }
 
 return LIP_SUCCESS;
}
