#include <mpi.h>
#include "lip.h"
#include <lipimpl.h>
#include <stdlib.h>
#include <stdio.h>
#include <correct.h>
#include "debug.h"
#define CD(j,k) (((j)-1)/(k)+1)
#define GHOST_AREA_RATE 2
/*
After calling this function all tab array elements [0..border-1] 
are not greater than elements [border..count-1].
Elements of tab2 array are ordered in the same way as tab.
*/

void quick_part(unsigned long * tab, int * tab2, int count, int border)
{
   int l, r, i, j, min_l;
   unsigned long med, temp;
   int temp_int;
   l=0;
   r = count-1;

   do
   {
      med = tab[(l+r)/2];
      i=l;
      j=r;
      while(i<j)
      { 
         while(tab[i]<med) i++;
         while(tab[j]>med) j--;
         if(i<j)
         {
            temp = tab[i];
            tab[i] = tab[j];
            tab[j] = temp;
            temp_int = tab2[i];
            tab2[i] = tab2[j];
            tab2[j] = temp_int;
            i++;
            j--;
         }
      }
      if(i==j)
      {
         if(tab[j]>med) j--;
         else i++;
      }
      /* now tab[0..j] <= tab[i..count-1] */
      if(i<border)
      {
         l = i;
      }
      else
      {
         r = j;
      }   
   }   while (l!=border&&r>l);
} /* end quick_part
*/

/* Fuction performs the same action as 
   LIP_create_hilbert_distribution_in_core
   except input data and output datamap 
   are out of core.
*/


int LIP_create_hilbert_distribution_ooc ( LIP_Datamap datamap,
                                          LIP_Datamap * new_datamap, 
                                          MPI_File XFile, 
                                          MPI_File YFile, 
                                          MPI_File ZFile, 
                                          int count,
                                          int page_size,
                                          int sample_count,
                                          void* vertex,
                                          double edge,
                                          double minedge)
{ 
  char * pname="LIP_create_hilbert_distribution_ooc";
  int pages;
  int i,j,l, res, itemp;
  int sample_index;
  double *coX, *coY, *coZ; /*the coordinates of the points
*/
  double * ver;
  double vertexX,vertexY,vertexZ, rr;
  double *samplex,*sampley,*samplez;
  int * displacements, * sizes, * index_buckets;
  int * procs; /*memory buffer for procss part od translation table
*/
  unsigned long * h_index, * h_index_all, * borders, * h_index_buckets;
  int * g_index; /*array for global indices
*/
  int rank=LIP_env.rank;
  int block_offset; /*offset of node disk block
*/
  int page_offset; /*current page offset counted from beginning of a block
*/
  int mem_count;
  MPI_Datatype view_datatype; 
  MPI_Info info;
  MPI_Status status;
  MPI_Offset mo;
  LIP_IDatamap* dm, *new_dm;
  MPI_File * h_bucket_files; /*buckets for hilbert indices
*/
  MPI_File * i_bucket_files; /*buckets for global indices
*/
  char temp_fname[100];
  unsigned long ultemp;
  int left_sum, left_disp, right_disp, bucket_size;
  int ret;
/*#ifdef DEBUG  
  LIP_DEBUG_create_file();
#endif*/
  if(1==LIP_env.size){
    *new_datamap=LIP_DATAMAP_NULL;
    return LIP_SUCCESS;
  }


  MPI_Info_create(&info);
  pages = CD(count, page_size);
  dm=LIP_Datamap_convert(datamap);
  if(NULL==dm){
   LIP_Error_quit(pname,"old datamap is invalid",LIP_INVALID_ARGUMENT);
  }
  
  if(dm->type!=LIP_DATAMAP_BLOCK){
   return LIP_ERROR;
  }

  LIP_Datamap_create(LIP_DATAMAP_INDIRECT,dm->global_size,&(dm->local_size),
                     LIP_INDEXTYPE_INT,1,LIP_NONLOCAL|LIP_OOC,new_datamap);

  new_dm=LIP_Datamap_convert(*new_datamap);
    if(NULL==new_dm){
   LIP_Error_quit(pname,"new datamap is invalid",LIP_INVALID_ARGUMENT);
  }

  new_dm->ttable_ooc_size=dm->node_size;
  new_dm->ttable_ic_size=2*page_size; /*the same size of in core sections as page size for partitioning 
*/
  new_dm->ttable_ic_real_size=page_size*3;/*plus ghost area*/
/*  imap->ttable_ic_size=ttable_membuf_size;
*/
 /* new_dm->ttable_ic_real_size=2*dm->node_size/4;
*/
  new_dm->procs=malloc(new_dm->ttable_ic_real_size*sizeof(int));
  LIP_Assert(new_dm->procs,pname,"allocating memory for new datamap",LIP_OUT_OF_MEMORY);
  new_dm->local_adr=malloc(new_dm->ttable_ic_real_size*sizeof(int));
  LIP_Assert(new_dm->local_adr,pname,"allocating memory for new datamap",LIP_OUT_OF_MEMORY);
  ret=LIP_IOBufmap_create(new_dm->ttable_ic_size,&(new_dm->tt_io_bmap));
  if(LIP_SUCCESS!=ret){
    return ret;
  }

  block_offset=(dm->local_size) * LIP_env.rank;
   
  ver=(double*)vertex;
  vertexX=ver[0];
  vertexY=ver[1];
  if (coZ!=NULL){vertexZ=ver[2];}

  procs = (int*) malloc (sizeof(int)*page_size);
  LIP_Assert(procs,pname,"allocating memory for procs",LIP_OUT_OF_MEMORY);
  /* allocate memory for input data */
  coX = (double*) malloc (sizeof(double)*page_size);
  LIP_Assert(coX,pname,"allocating memory for coX",LIP_OUT_OF_MEMORY);
  coY = (double*) malloc (sizeof(double)*page_size);
  LIP_Assert(coY,pname,"allocating memory for coY",LIP_OUT_OF_MEMORY);

  if (ZFile==MPI_FILE_NULL) coZ = NULL;
  else 
  {
       coZ = (double*) malloc (sizeof(double)*page_size);
       LIP_Assert(coZ,pname,"allocating memory for coZ",LIP_OUT_OF_MEMORY);
  }
  /***** choose sample *****/

  h_index=(unsigned long*)(malloc(sizeof(unsigned long)*sample_count));
  LIP_Assert(h_index,pname,"allocating memory for hindex",LIP_OUT_OF_MEMORY);
  samplex=(double*)(malloc(sizeof(double)*sample_count));
  LIP_Assert(samplex,pname,"allocating memory for samplex",LIP_OUT_OF_MEMORY);
  sampley=(double*)(malloc(sizeof(double)*sample_count));
  LIP_Assert(sampley,pname,"allocating memory for sampley",LIP_OUT_OF_MEMORY);

  if(coZ!=NULL)
  { 
    samplez=(double*)(malloc(sizeof(double)*sample_count));
    LIP_Assert(samplez,pname,"allocating memory for procs",LIP_OUT_OF_MEMORY);

  }
  displacements = (int *) malloc(sizeof(int)*sample_count);
  LIP_Assert(displacements,pname,"allocating memory for displacements",LIP_OUT_OF_MEMORY);

  srand(1);
  for(i=0;i<sample_count;i++)
  {
    rr=rand();
    sample_index=(int)block_offset+(dm->local_size*rr/(RAND_MAX+1.0));
    displacements[i]=sample_index;
  }
  /* sort random displacememnts, as required by MPI_File_set_view */
  qsort(displacements, sample_count, sizeof(int),sortint);
#ifdef DEBUG 
  fprintf(debf,"wybieramy losowe pozycje:\n");
  for(i=0;i<sample_count;i++)
  {
    fprintf (debf,"%u ",displacements[i]);
  }
  fprintf(debf,"\n");
#endif  
  
  /* create random view to read sample from files */

  LIP_Type_create_indexed_block(sample_count, 1 ,displacements, MPI_DOUBLE, &view_datatype);
  MPI_Type_commit(&view_datatype);  
  MPI_File_set_view(XFile, 0, MPI_DOUBLE, view_datatype,"native", MPI_INFO_NULL);
#ifdef DEBUG   
  fprintf(debf,"set view OK, sample_count = %i\n", sample_count); 
#endif 
  MPI_File_read_at(XFile, 0, samplex, sample_count, MPI_DOUBLE, &status);
#ifdef DEBUG  
  fprintf(debf,"read OK\n"); 
#endif 
  MPI_File_set_view(YFile, 0, MPI_DOUBLE, view_datatype,"native", MPI_INFO_NULL);
  MPI_File_read_at(YFile, 0, sampley, sample_count, MPI_DOUBLE, &status);
  if(coZ!=NULL)
  {
    MPI_File_set_view(ZFile, 0, MPI_DOUBLE, view_datatype,"native", info);
    MPI_File_read_at(ZFile, 0, samplez, sample_count, MPI_DOUBLE, &status);
  } 
  MPI_Type_free(&view_datatype);

#ifdef DEBUG
  fprintf(debf,"Wspolrzedne\n");
  for(i=0;i<sample_count;i++){ 
    fprintf(debf,"%1.1f ",samplex[i]);
  } 
  fprintf(debf,"\n"); 
  
  
  for(i=0;i<sample_count;i++){
    fprintf(debf,"%1.1f ",sampley[i]);
  } 
  fprintf(debf,"\n"); 
  if(coZ!=NULL){
   for(i=0;i<sample_count;i++){
    fprintf(debf,"%1.1f ",samplez[i]);
   }
   
   fprintf(debf,"\n"); 
  }
#endif  

  /*count hilbert index
*/
  if (coZ!=NULL)
  {
    hilbert3d(sample_count,samplex,sampley,samplez,vertexX,vertexY,vertexZ,edge,minedge,
                  h_index);
    free(samplez);
  }
  else
  {
     hilbert2d(sample_count,samplex,sampley,vertexX,vertexY,edge,minedge,
                    h_index);
  }
  free(samplex);
  free(sampley);
#ifdef DEBUG  
  fprintf(debf,"Indeksy Hilberta:\n");
  for(i=0;i<sample_count;i++)
  {
    fprintf(debf,"%u ",h_index[i]);
  } 
  fprintf(debf,"\n"); 
#endif  
  /*sorting hilbert indices 
*/
  qsort(h_index,sample_count,sizeof(unsigned long), (int (*)(const void *, const void *))sortLD);
 
#ifdef DEBUG 
  fprintf(debf,"Posortowane indeksy Hilberta\n");
  for(i=0;i<sample_count;i++)
  {
    fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
#endif
  
  /*merge indices
*/
  
  h_index_all=chain_merge(h_index,sample_count,MPI_UNSIGNED_LONG);
  
   
  if(LIP_env.rank==0)
  {
#ifdef DEBUG
    fprintf(debf,"Posortowane indeksy Hilberta po scaleniu\n");
  
    for(i=0;i<sample_count*LIP_env.size;i++)
    {
      fprintf (debf,"%u ",h_index_all[i]);
    }
    fprintf(debf,"\n");
#endif

    /*count borders
*/
    borders=malloc(sizeof(unsigned long)*(LIP_env.size));
    LIP_Assert(borders,pname,"allocating memory for borders",LIP_OUT_OF_MEMORY);

    borders[0]=0;
    for(i=1;i<LIP_env.size;i++)
    {
      borders[i]=h_index_all[i*sample_count];
    }
   
#ifdef DEBUG   
    for(i=0;i<LIP_env.size;i++)
    {
      fprintf (debf,"%u ",borders[i]);
    }
    fprintf(debf,"\n");
#endif

  }
  else /* rank!=0
*/
  {
     borders=malloc(sizeof(unsigned long)*(LIP_env.size));
     LIP_Assert(borders,pname,"allocating memory for borders",LIP_OUT_OF_MEMORY);

  }
  free (h_index_all);

  MPI_Bcast(borders,LIP_env.size,MPI_UNSIGNED_LONG,0,LIP_env.comm);
#ifdef DEBUG
  fprintf(debf,"granice:\n");
  for(i=0;i<LIP_env.size;i++)
  {
    fprintf (debf,"%u ",borders[i]);
  }
  fprintf(debf,"\n");
#endif  

  /***** Main bucket sorting loop *****/

  page_offset = 0;
  mem_count = page_size;
  h_index = realloc(h_index,sizeof(unsigned long)*page_size);
  LIP_Assert(h_index,pname,"allocating memory for h_index",LIP_OUT_OF_MEMORY);
  sizes = (int*) malloc(LIP_env.size*sizeof(int));
  LIP_Assert(sizes,pname,"allocating memory for sizes",LIP_OUT_OF_MEMORY);
  displacements = (int*) realloc(displacements, LIP_env.size*sizeof(int));
  LIP_Assert(displacements,pname,"allocating memory for displacements",LIP_OUT_OF_MEMORY);
  h_index_buckets=malloc(sizeof(unsigned long)*(page_size));
  LIP_Assert(h_index_buckets,pname,"allocating memory for h_index_buckets",LIP_OUT_OF_MEMORY);
  index_buckets=malloc(sizeof(int)*(page_size));
  LIP_Assert(index_buckets,pname,"allocating memory for buckets",LIP_OUT_OF_MEMORY);
  MPI_File_set_view(XFile,0,MPI_DOUBLE, MPI_DOUBLE, "native", info);
  MPI_File_set_view(YFile,0,MPI_DOUBLE, MPI_DOUBLE, "native", info);
  if (coZ!=NULL) MPI_File_set_view(ZFile,0,MPI_DOUBLE, MPI_DOUBLE, "native", info);
  /* create bucket files array*/
  h_bucket_files = (MPI_File*) malloc(sizeof(MPI_File)*LIP_env.size);
  i_bucket_files = (MPI_File*) malloc(sizeof(MPI_File)*LIP_env.size);
  for(i=0;i<LIP_env.size; i++)
  {
    sprintf(temp_fname, "_LIP_temp_h_bucket%i", i);
    MPI_File_open(LIP_env.comm, temp_fname, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &h_bucket_files[i]);
    MPI_File_set_view(h_bucket_files[i],0,MPI_UNSIGNED_LONG, MPI_UNSIGNED_LONG, "native", info);
    sprintf(temp_fname, "_LIP_temp_i_bucket%i", i);
    MPI_File_open(LIP_env.comm, temp_fname, MPI_MODE_RDWR |
                MPI_MODE_CREATE, MPI_INFO_NULL, &i_bucket_files[i]);
    MPI_File_set_view(i_bucket_files[i],0,MPI_INT, MPI_INT, "native", info);
  }
  MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT, "native", info);  
  while(page_offset<count) /*main loop thru all data on disk
*/
  {
    /* check if not enough data to fill the page */
    if(count-page_offset<mem_count) mem_count = count-page_offset;
    /* read page from disk */
    MPI_File_read_at(XFile, block_offset+page_offset, coX, mem_count, MPI_DOUBLE, &status);
    MPI_File_read_at(YFile, block_offset+page_offset, coY, mem_count, MPI_DOUBLE, &status);  
    if (coZ!=NULL) 
      MPI_File_read_at(ZFile, block_offset+page_offset, coZ, mem_count, MPI_DOUBLE, &status);
    for(i=0;i<LIP_env.size;i++)  
    {
      sizes[i]=0;
    }
    /* compute Hilbert indices for all array elements */
    if (coZ!=NULL) hilbert3d(mem_count, coX, coY, coZ, vertexX, vertexY, vertexZ, edge, minedge, h_index);
    else hilbert2d(mem_count, coX, coY, vertexX, vertexY, edge, minedge, h_index);
    /* assign bucket number to each index */
    for(i=0;i<mem_count;i++)
    {
      l=0;
      while (borders[l]<=h_index[i])
      {
        l++; 
        if (LIP_env.size==l) break;
      }
      procs[i]=l-1;
      sizes[procs[i]]++;  
    }
#ifdef DEBUG
    fprintf(debf,"indeksy Hilberta:\n");
    for(i=0;i<mem_count;i++){
        fprintf (debf,"%u ",h_index[i]);
    }
    fprintf(debf,"\n");

    fprintf(debf,"naleza do procesorow:\n");
    for(i=0;i<mem_count;i++){
      fprintf (debf,"%i ",procs[i]);
    }
    fprintf(debf,"\n");
#endif
    /* compute places for buckets */
    displacements[0]=0;
    for(i=1;i<LIP_env.size;i++)
    {
      displacements[i]=displacements[i-1]+sizes[i-1];
      sizes[i-1]=0;
    }
    sizes[LIP_env.size-1]=0;   
    
    /* move indices to buckets */
    j=0;
    for(i=0;i<mem_count;i++)
    {
      h_index_buckets[displacements[procs[i]]+sizes[procs[i]]]=h_index[i];
      index_buckets[displacements[procs[i]]+sizes[procs[i]]]=i+block_offset+page_offset; /*global index
*/
      sizes[procs[i]]++;
    }
#ifdef DEBUG
    fprintf(debf,"indeksy Hilberta w kubelkach:\n");
    for(i=0;i<mem_count;i++){
       fprintf (debf,"%u ",h_index_buckets[i]);
    }
    fprintf(debf,"\n");
#endif
     
    /* write buckets to disk */
    for(i=0; i<LIP_env.size; i++)
    {
      MPI_File_write_ordered(h_bucket_files[i], &h_index_buckets[displacements[i]], 
                            sizes[i], MPI_UNSIGNED_LONG, &status);
      MPI_File_write_ordered(i_bucket_files[i], &index_buckets[displacements[i]], 
                            sizes[i], MPI_INT, &status);
    }
    /* write part of translation table */
    MPI_File_write_at(new_dm->procs_file, block_offset+page_offset, procs, mem_count, MPI_INT, &status);
    page_offset+=mem_count;
  } /*end of main bucket sorting loop
*/
  free (h_index_buckets);
  h_index_buckets = NULL;
  free (index_buckets);
  index_buckets = NULL;
  free (coX);
  coX = NULL;
  free (coY);
  coY = NULL;
  free (borders);
  borders = NULL;
  if(coZ!=NULL)
  {
    free (coZ);
    coZ = NULL;
  }
  MPI_Barrier(LIP_env.comm); /*wait for all shared file operations 
*/

#ifdef DEBUG
  MPI_File_get_position_shared(h_bucket_files[LIP_env.rank], &j);
  fprintf(debf, "rozmiar pliku z kubelkiem: %i\n", j);
  fprintf(debf,"zawartosc pliku mojego kubelka - indeksy Hilberta:\n");
  for(i=0;i<j;i++){
     MPI_File_read(h_bucket_files[LIP_env.rank], &ultemp, 1, MPI_UNSIGNED_LONG, &status); 
     fprintf (debf,"%u ", ultemp);
  }
  fprintf(debf,"\n");
  fprintf(debf,"zawartosc pliku mojego kubelka - indeksy globalne:\n");
  for(i=0;i<j;i++){
     MPI_File_read(i_bucket_files[LIP_env.rank], &ultemp, 1, MPI_INT, &status); 
     fprintf (debf,"%u ", ultemp);
  }
  fprintf(debf,"\n");
#endif

#ifdef DEBUG
  MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT, "native", info);  
  fprintf(debf,"zawartosc fragmentu tablicy translacji:\n");
  for(i=0;i<dm->local_size;i++){
     MPI_File_read_at(new_dm->procs_file, i+block_offset, &itemp, 1, MPI_INT, &status); 
     fprintf (debf,"%u ", itemp);
  }
  fprintf(debf,"\n");
#endif

  /***** correction *****/

  /* find difference between current and correct bucket size */
  left_sum = 0;
  for(i=0; i<LIP_env.rank; i++)
  {
    MPI_File_get_position_shared(h_bucket_files[i], &mo);
    j=(int) mo;
    left_sum += j;
  }
  left_disp = block_offset - left_sum;
  MPI_File_get_position_shared(h_bucket_files[LIP_env.rank], &mo);
   bucket_size=(int) mo;
  right_disp = -left_disp + bucket_size - dm->local_size;
#ifdef DEBUG
  fprintf(debf, "w kuble powinno byc: %i\n", dm->local_size);
  fprintf(debf, "z lewej trzeba oddac: %i\n", left_disp);
  fprintf(debf, "z prawej trzeba oddac: %i\n", right_disp);
  fprintf(debf, "kubelek ma nadmiar: %i elementow\n", right_disp+left_disp);  
#endif
  /* we care only for our extra indices */
  if (left_disp<0) left_disp = 0;
  if (right_disp<0) right_disp = 0;
  if (left_disp>bucket_size || bucket_size-right_disp<0 || left_disp>page_size || right_disp>page_size)
  {
    /* here we crash - please try better random sample */
    /*FIXME: close all temp files and free memory
*/
#ifdef DEBUG
    fprintf(debf, "zbyt duze rozbieznosci - nie potrafimy skorygowac: %i\n", dm->local_size);
#endif
    printf ("zbyt duze rozbieznosci - nie potrafimy skorygowac\n");
    LIP_Error_quit(pname, "Sample size too small to generate distribution", LIP_ERROR);
    return LIP_ERROR;
  }
  
  page_offset = 0;
  mem_count = page_size;
  h_index = (unsigned long *) realloc(h_index, (page_size+right_disp+left_disp)
                                                 * sizeof(unsigned long));
  g_index = (int *) malloc((page_size+right_disp+left_disp)
                                                 * sizeof(int));
  /* read first page */
  if(right_disp+left_disp>0)
  {
    if(bucket_size-page_offset<mem_count) 
    {
      mem_count = bucket_size-page_offset;
    }
    MPI_File_read_at(h_bucket_files[LIP_env.rank], page_offset, h_index+left_disp, mem_count, MPI_UNSIGNED_LONG, &status);
    MPI_File_read_at(i_bucket_files[LIP_env.rank], page_offset, g_index+left_disp, mem_count, MPI_INT, &status);
    if(left_disp>0) 
    {
      quick_part(h_index+left_disp, g_index+left_disp, mem_count, left_disp);
      for(i=0; i<left_disp; i++) 
      {
        h_index[i] = h_index[left_disp+i];
        g_index[i] = g_index[left_disp+i];
      }
    }
    if(right_disp>0)
    { 
      quick_part(h_index+left_disp+left_disp, g_index+left_disp+left_disp, mem_count-left_disp, mem_count-left_disp-right_disp); 
      for(i=0; i<right_disp; i++) 
      {
        h_index[i+left_disp+mem_count] = h_index[i+left_disp+mem_count-right_disp];    
        g_index[i+left_disp+mem_count] = g_index[i+left_disp+mem_count-right_disp];    
      }
    }
    page_offset+=mem_count;
#ifdef DEBUG
  fprintf(debf,"h_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"g_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",g_index[i]);
  }
  fprintf(debf,"\n");

#endif
  }

  while(page_offset<bucket_size) /*loop thru all pages
*/
  {
    if(bucket_size-page_offset<mem_count) 
    {
      mem_count = bucket_size-page_offset;
      for(i=0; i<right_disp; i++)
      {
        h_index[i+mem_count+left_disp] = h_index[i+page_size+left_disp];
        g_index[i+mem_count+left_disp] = g_index[i+page_size+left_disp];
      }
    }
    MPI_File_read_at(h_bucket_files[LIP_env.rank], page_offset, h_index+left_disp, mem_count, MPI_UNSIGNED_LONG, &status);
    MPI_File_read_at(i_bucket_files[LIP_env.rank], page_offset, g_index+left_disp, mem_count, MPI_INT, &status);

#ifdef DEBUG
  fprintf(debf,"h_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"g_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",g_index[i]);
  }
  fprintf(debf,"\n");

#endif

    if(left_disp>0) quick_part(h_index, g_index, mem_count+right_disp+left_disp, left_disp);
    if(right_disp>0) quick_part(h_index+left_disp, g_index+left_disp, mem_count+right_disp, mem_count);     
    page_offset+=mem_count;
#ifdef DEBUG
  fprintf(debf,"h_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"g_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",g_index[i]);
  }
  fprintf(debf,"\n");

#endif

  }  
#ifdef DEBUG
  fprintf(debf,"h_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"g_index:\n");
  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",g_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"w lewo przerzucam:\n");
  for(i=0;i<left_disp;i++){
     fprintf (debf,"%u ",h_index[i]);
  }
  fprintf(debf,"\n");
  fprintf(debf,"w prawo przerzucam:\n");
  for(i=0;i<right_disp;i++){
     fprintf (debf,"%u ",h_index[i+mem_count+left_disp]);
  }
  fprintf(debf,"\n");
#endif
  /* write correction info to translation table */
  if(left_disp>0)
  {
    qsort(g_index, left_disp, sizeof(int), sortint);
    
/*  for(i=0;i<left_disp+right_disp+page_size;i++){
     fprintf (debf,"%u ",g_index[i]);
  }*/
    LIP_Type_create_indexed_block(left_disp, 1 ,g_index, MPI_INT, &view_datatype);
    MPI_Type_commit(&view_datatype);  
    MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, view_datatype,"native", MPI_INFO_NULL);
    /*MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT,"native", info);
*/
    for(i=0; i<left_disp; i++) procs[i]=LIP_env.rank-1;
    MPI_File_write_at(new_dm->procs_file, 0, procs, left_disp, MPI_INT, &status);
    /* printf("tralalalaal %i\n", procs[1]);
*/
    MPI_Type_free(&view_datatype);
  }
  else MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT,"native", info);
  if(right_disp>0)
  {
    qsort(g_index+left_disp+mem_count, right_disp, sizeof(int), sortint);
    LIP_Type_create_indexed_block(right_disp+1, 1 ,g_index+left_disp+mem_count, MPI_INT, &view_datatype);
    res = MPI_Type_commit(&view_datatype);  
    MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, view_datatype,"native", info);
    for(i=0; i<right_disp; i++) procs[i]=LIP_env.rank+1;
    MPI_File_write_at(new_dm->procs_file, 0, procs, right_disp, MPI_INT, &status);
    MPI_Type_free(&view_datatype);
  }
  else MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT,"native", info);
#ifdef DEBUG
  MPI_Barrier(LIP_env.comm); /*wait for all shared file operations 
*/
  MPI_File_set_view(new_dm->procs_file, 0, MPI_INT, MPI_INT, "native", info);  
  fprintf(debf,"zawartosc fragmentu tablicy translacji po korekcji:\n");
  for(i=0;i<dm->local_size;i++){
     MPI_File_read_at(new_dm->procs_file, i+block_offset, &itemp, 1, MPI_INT, &status); 
     fprintf (debf,"%i ", itemp);
  }
  fprintf(debf,"\n");
#endif


  for(i=0;i<LIP_env.size; i++)
  {
    sprintf(temp_fname, "_LIP_temp_h_bucket%i", i);
    MPI_File_close (&h_bucket_files[i]);
    if (rank==0) MPI_File_delete(temp_fname, info);
    sprintf(temp_fname, "_LIP_temp_i_bucket%i", i);
    MPI_File_close (&i_bucket_files[i]);
    if (rank==0) MPI_File_delete(temp_fname, info);
  }
  /* free memory */
  
  free (procs);

  free (displacements);
  if(g_index!=NULL)
    free(g_index);
  if(h_index!=NULL)
    free(h_index);
}
