#include <mpi.h>
#include "lip.h"
#include <lipimpl.h>
#include <stdlib.h>
#include <stdio.h>
#include "correct.h"
#include "debug.h"

/*
  Function remaps data according to new irregular partition
  stored in new_datamap object. It must contain
  procs_file generated properly by LIP_create_hilbert_distribution_ooc
  or any other partitioner.
  If flag LIP_CREATE_LOCAL is specified second part 
  of translation table is created: local indices of data points
  are calculated according to new distribution and stored in new datamap.
  IF flag LIP_MOVE_DATA is specified then data stored in
  data_file is rearranged according to new distribution
  and stored in this new ordre in data_out_file.
*/
  
int LIP_remap_ooc (LIP_Datamap datamap,
                   LIP_Datamap * new_datamap,
                   MPI_File data_file,
                   MPI_File data_out_file, 
                   int count, 
                   int page_size,
                   MPI_Datatype datatype,
                   int flags)
{
     char * pname="LIP_remap_ooc";
    int * procs; /*memory buffer for procss part od translation table*/
    int rank=LIP_env.rank;
    int block_offset; /*offset of node disk block*/
    int page_offset; /*current page offset counted from beginning of a block*/
    int mem_count;
    int * out_page_offsets, * out_offsets;/* array of offsets in output file*/
    int * esizes, * edispls, * elocations;
    MPI_Datatype view_datatype, data_filetype,datatype_for_mov; 
    MPI_Info info;
    MPI_Status status;
    MPI_File procs_file, local_ind_file;
    LIP_IDatamap* dm, *new_dm;
    int i,j, ret;
    void * data_buffer;
    double dtemp;
    int f_local, f_move; /* flags*/
    int type_size;
    int rest;
    MPI_Info_create(&info);
    
    if(1==LIP_env.size){
        rest=count;
    
        MPI_Type_size(datatype,&type_size);
        data_buffer=malloc(page_size*type_size);
        LIP_Assert(data_buffer,pname,"not enough memory for allocating data for moving",
	           LIP_OUT_OF_MEMORY);
	
	MPI_Type_contiguous(page_size,datatype,&datatype_for_mov);
	MPI_Type_commit(&datatype_for_mov);
	MPI_File_set_view(data_file, 0, datatype, datatype_for_mov, "native", info);    
	MPI_File_set_view(data_out_file, 0, datatype, datatype_for_mov, "native", info);    
	
	while(rest>page_size){
    	    MPI_File_read_at(data_file, count-rest, data_buffer, 1,
datatype_for_mov, &status); 
	    
	    MPI_File_write_at(data_out_file, 
          		count-rest,data_buffer,1, datatype_for_mov,
&status);
            rest-=page_size;
    	
	}
	MPI_Type_free(&datatype_for_mov);
        
	if(rest!=0){
	MPI_Type_contiguous(rest,datatype,&datatype_for_mov);
	MPI_Type_commit(&datatype_for_mov);
	MPI_File_set_view(data_file, 0, datatype, datatype_for_mov, "native", info);    
	MPI_File_set_view(data_out_file, 0, datatype, datatype_for_mov, "native", info);    
	
    	MPI_File_read_at(data_file, count-rest, data_buffer, 1,
datatype_for_mov, &status); 
        MPI_File_write_at(data_out_file, 
          		count-rest,data_buffer,1, datatype_for_mov,
&status);
	
	MPI_Type_free(&datatype_for_mov);
        }        
	free(data_buffer);
	return LIP_SUCCESS;
    }
    
    
     
    /* check flags */
    f_local = flags & LIP_CREATE_LOCAL;
    f_move = flags & LIP_MOVE_DATA;
    /* allcate memory buffers */
    procs = (int*) malloc (sizeof(int)*page_size);
    LIP_Assert(procs,pname,"allocating procs array",LIP_OUT_OF_MEMORY);
    elocations = (int*) malloc (sizeof(int)*page_size);
    LIP_Assert(elocations,pname,"allocating elocations array",LIP_OUT_OF_MEMORY);
    esizes = malloc( (sizeof *esizes) * LIP_env.size );
    LIP_Assert(esizes,pname,"allocating esizes array",LIP_OUT_OF_MEMORY);
    edispls = malloc( (sizeof *edispls) * LIP_env.size );
    LIP_Assert(edispls,pname,"allocating edispls array",LIP_OUT_OF_MEMORY);
    out_page_offsets = malloc( (sizeof *out_page_offsets) * LIP_env.size );
    LIP_Assert(out_page_offsets,pname,"allocating out_page_offsets array",LIP_OUT_OF_MEMORY);
    out_offsets = malloc( (sizeof *out_offsets) * LIP_env.size );
    LIP_Assert(out_offsets,pname,"allocating out_offsets array",LIP_OUT_OF_MEMORY);
    MPI_Type_size(datatype, &type_size);
    if (f_move) data_buffer = malloc ( type_size*page_size);
    dm=LIP_Datamap_convert(datamap);
    if(NULL==dm){
        LIP_Error_quit(pname,"old datamap is invalid",LIP_INVALID_ARGUMENT);
    }
    if (dm->type!=LIP_DATAMAP_BLOCK){
     return LIP_ERROR;
    }
    new_dm=LIP_Datamap_convert(* new_datamap);
    if(NULL==new_dm){
        LIP_Error_quit(pname,"new datamap is invalid",LIP_INVALID_ARGUMENT);
    }
    block_offset=(dm->local_size) * rank;
    /*here we operate on files in new datamap*/
    procs_file = new_dm->procs_file;
    local_ind_file = new_dm->local_adr_file;
    page_offset = 0;
    mem_count = page_size;
    ret = 0;
    
#ifdef DEBUG
    fprintf (debf, "setting view \n");
#endif
 
    MPI_File_set_view(procs_file, 0, MPI_INT, MPI_INT, "native", info);
    if (f_move) MPI_File_set_view(data_out_file, 0, datatype, datatype, "native", info);    

#ifdef DEBUG
    fprintf (debf, "view set \n");
#endif

    for (i=0; i<LIP_env.size; i++) out_offsets[i] = 0;            
    while(page_offset<count) /*main loop thru all data on disk*/
    {
        /* check if not enough data to fill the page */
        if(count-page_offset<mem_count) mem_count = count-page_offset;
        /* read page from disk */
        MPI_File_read_at(procs_file, block_offset+page_offset, 
                         procs, mem_count, MPI_INT, &status);
                         

#ifdef DEBUG
        fprintf (debf, "procs: \n");   
        for(i=0;i<mem_count;i++)
        {
          fprintf (debf,"%u ",procs[i]);
        }
        fprintf(debf,"\n");
#endif                          
                         
        for (i=0; i<LIP_env.size; i++) esizes[i] = edispls[i] = 0;
        for (i=0; i<mem_count; i++) esizes[procs[i]]++;
        for (i=1; i<LIP_env.size; i++) edispls[i] = edispls[i-1]+esizes[i-1];
#ifdef DEBUG
        fprintf (debf, "esizes: \n");   
        for(i=0;i<LIP_env.size;i++)
        {
          fprintf (debf,"%u ",esizes[i]);
        }
        fprintf(debf,"\n");
        fprintf (debf, "edispls: \n");   
        for(i=0;i<LIP_env.size;i++)
        {
          fprintf (debf,"%u ",edispls[i]);
        }
        fprintf(debf,"\n");
        fprintf (debf,"mem_count%u ",mem_count);
        
#endif        
        /* compute offsets for this page */
        MPI_Scan(esizes, out_page_offsets, LIP_env.size, MPI_INT, MPI_SUM, LIP_env.comm);
        for (i=0; i<LIP_env.size; i++) out_page_offsets[i] -= esizes[i];

#ifdef DEBUG
        fprintf (debf, "out_page_offsets: \n");   
        for(i=0;i<LIP_env.size;i++)
        {
          fprintf (debf,"%u ",out_page_offsets[i]);
        }
        fprintf(debf,"\n");
#endif       
        for(i=0; i<LIP_env.size; i++) esizes[i] = 0;
        for(i=0; i<mem_count; i++)
        {
            elocations[edispls[procs[i]] + esizes[procs[i]]] = i;
            esizes[ procs[i] ]++;
        }         
        for (j=0; j<LIP_env.size; j++) /*process each part of elocations array*/
        {
            /* create datatype for local_ind_file */
            if (esizes[j]!=0)
            {
                if (f_local) 
                {
                     LIP_Type_create_indexed_block
                              (esizes[j], 1 , &elocations[edispls[j]], MPI_INT, 
                               &view_datatype);
                     MPI_Type_commit(&view_datatype);
                }
                if(f_move)
                {         
                     LIP_Type_create_indexed_block(esizes[j], 1 , 
		               &elocations[edispls[j]], datatype, &data_filetype); 
                     MPI_Type_commit(&data_filetype);          
                }
                if (f_local) MPI_File_set_view(local_ind_file, 
		                         sizeof(int)*(block_offset+page_offset), MPI_INT, 
		                         view_datatype,"native", MPI_INFO_NULL);
                for(i=0; i<esizes[j]; i++) elocations[i+edispls[j]] = 
		                         i+out_offsets[j]+out_page_offsets[j];
                /* write data to local_ind_file */
                if (f_local) MPI_File_write_at(local_ind_file, 0, 
          		&elocations[edispls[j]], esizes[j], MPI_INT, &status);
                /* copy data from old file to new file */
                if (f_move)
                { 
                          MPI_File_set_view(data_file, type_size*
          		          (block_offset+page_offset), datatype, data_filetype,"native", MPI_INFO_NULL);
                          MPI_File_read_at(data_file, 0, data_buffer, esizes[j], 
          		           datatype, &status);
                }    
#ifdef DEBUG
if(f_move)
{
        fprintf (debf, "buffer: \n");   
        for(i=0;i<esizes[j];i++)
        {
          fprintf (debf,"%2.2f ",((double *) data_buffer)[i]);
        }
        fprintf(debf,"\n");
}
#endif  
                    
                    
                if (f_move) MPI_File_write_at(data_out_file, 
          		          dm->local_size*j+out_offsets[j]+out_page_offsets[j], 
          		          data_buffer, esizes[j], datatype, &status);

            }
            else 
            {
                if (f_local) MPI_File_set_view(local_ind_file, sizeof(int)*(block_offset*rank+page_offset), MPI_INT, MPI_INT,"native", MPI_INFO_NULL);
                if (f_move) MPI_File_set_view(data_file, sizeof(double)*(block_offset+page_offset), datatype, datatype,"native", MPI_INFO_NULL);
            }
        }
#ifdef DEBUG
        fprintf (debf, "elocations: \n");   
        for(i=0;i<mem_count;i++)
        {
          fprintf (debf,"%u ",elocations[i]);
        }
        fprintf(debf,"\n");
#endif               
        

        /* get total sizes for this page */
        if(rank==LIP_env.size-1) 
        {
            for (i=0; i<LIP_env.size; i++) esizes[i]+=out_page_offsets[i];
        }
        MPI_Bcast(esizes, LIP_env.size, MPI_INT, LIP_env.size-1, LIP_env.comm);
        /* update out_offsets array */
        for (i=0; i<LIP_env.size; i++) out_offsets[i]+=esizes[i];
        page_offset+=mem_count;
    } /*end of main loop*/
    
#ifdef DEBUG
  MPI_Barrier(LIP_env.comm); /*wait for all operations*/
  if (f_local)
  { 
       ret =  MPI_File_set_view(local_ind_file, 0, MPI_INT, MPI_INT, "native", info);  
      /* printf ("%i ", ret);*/
       fprintf(debf,"zawartosc czesci local_ind tablicy translacji:\n");
       for(i=0;i<dm->local_size;i++){
          ret = MPI_File_read_at(local_ind_file, i+block_offset, &j, 1, MPI_INT, &status); 
          /*printf ("%i ", ret);*/
          fprintf (debf,"%i ", j);
       }
       fprintf(debf,"\n");
  }
#endif  



#ifdef DEBUG
if (f_move)
{
  MPI_Barrier(LIP_env.comm); /*wait for all operations */
  ret =  MPI_File_set_view(data_out_file, 0, datatype, datatype, "native", info);  
  /*printf ("%i ", ret);*/
  fprintf(debf,"zawartosc calosci pliku z danymi:\n");
  for(i=0;i<dm->global_size;i++){
     ret = MPI_File_read_at(data_file, i, &dtemp, 1, datatype, &status); 
     /*printf ("%i ", ret);*/
     fprintf (debf,"%2.2f ", dtemp);
  }
  fprintf(debf,"\n");
}
#endif   


#ifdef DEBUG
if (f_move)
{
  MPI_Barrier(LIP_env.comm); /*wait for all operations */
  ret =  MPI_File_set_view(data_out_file, 0, datatype, datatype, "native", info);  
  /*printf ("%i ", ret);*/
  fprintf(debf,"zawartosc calosci pliku z przesunietymi danymi:\n");
  for(i=0;i<dm->global_size;i++){
     ret = MPI_File_read_at(data_out_file, i, &dtemp, 1, datatype, &status); 
     /*printf ("%i ", ret);*/
     fprintf (debf,"%2.2f ", dtemp);
  }
  fprintf(debf,"\n");
}
#endif   

    free (procs);
    free (elocations);
    free (esizes);
    free (edispls);
    free (out_page_offsets);
    free (out_offsets);
    if (f_move) free (data_buffer);

  return LIP_SUCCESS;
}

